# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import itertools
import numpy
from ._op import OpRun
from .op_max_pool_ import MaxPoolFloat, MaxPoolDouble  # pylint: disable=E0611,E0401


def _pool_get_output_shape(auto_pad, input_spatial_shape, kernel_spatial_shape,
                           strides_spatial):
    out_shape = [0] * len(input_spatial_shape)
    if auto_pad in (b'SAME_UPPER', b'SAME_LOWER'):
        for i in range(len(input_spatial_shape)):  # pylint: disable=C0200
            out_shape[i] = int(
                numpy.ceil(
                    float(input_spatial_shape[i]) / float(strides_spatial[i])))
    elif auto_pad == b'VALID':
        for i in range(len(input_spatial_shape)):  # pylint: disable=C0200
            out_shape[i] = int(
                numpy.ceil(float(input_spatial_shape[i] - (kernel_spatial_shape[i] - 1)) /
                           float(strides_spatial[i])))
    return out_shape


def _pool_impl(padded, x_shape, kernel_shape, strides_shape,
               out_shape, pad_shape, pooling_type,
               count_include_pad=0):
    spatial_size = len(x_shape) - 2
    y = numpy.zeros([x_shape[0], x_shape[1]] + list(out_shape))

    for shape in itertools.product(
            range(x_shape[0]), range(x_shape[1]),
            *[range(int((x_shape[i + 2] + pad_shape[i] - kernel_shape[i]) /
                        strides_shape[i] + 1))
              for i in range(spatial_size)]):
        window = padded[shape[0], shape[1]]
        window_vals = numpy.array(
            [window[i] for i in list(
                itertools.product(
                    *[range(strides_shape[i] * shape[i + 2],
                            strides_shape[i] * shape[i + 2] + kernel_shape[i])
                      for i in range(spatial_size)]))])
        if pooling_type == b'AVG':
            f = numpy.average
        elif pooling_type == b'MAX':
            f = numpy.max
        else:
            raise NotImplementedError(  # pragma: no cover
                f"Pooling type '{pooling_type}' does not support. Should be AVG, MAX.")

        if count_include_pad == 1 and pooling_type == b'AVG':
            y[shape] = f(window_vals)
        else:
            y[shape] = f(window_vals[numpy.where(~numpy.isnan(window_vals))])
    return y.astype(numpy.float32)


class MaxPool(OpRun):

    atts = {'auto_pad': b'NOTSET', 'ceil_mode': 0, 'dilations': [],
            'kernel_shape': [], 'pads': [], 'storage_order': 0,
            'strides': []}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=MaxPool.atts,
                       **options)
        self.auto_pad_ = self.auto_pad.decode('ascii')
        self.nb_outputs = len(onnx_node.output)
        self._init()

    def _init(self):
        self.rt32_ = MaxPoolFloat()
        self.rt64_ = MaxPoolDouble()
        for rt in [self.rt32_, self.rt64_]:
            rt.init(self.auto_pad,
                    numpy.array(self.dilations, dtype=numpy.int64),
                    self.ceil_mode,
                    self.storage_order,
                    numpy.array(self.kernel_shape, dtype=numpy.int64),
                    numpy.array(self.pads, dtype=numpy.int64),
                    numpy.array(self.strides, dtype=numpy.int64))

    def _run(self, X, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if X.dtype == numpy.float32:
            res = self.rt32_.compute(X)
        else:
            res = self.rt64_.compute(X)
        if self.nb_outputs == 1:
            return res[:1]
        return res



"""
@file
@brief Main functions decomposing einsum computation into
more simple functions.
"""
from itertools import permutations
import time
import math
import numpy
from onnx import helper
from ...onnx_tools.onnx2py_helper import guess_proto_dtype
from ...onnxrt.onnx_micro_runtime import OnnxMicroRuntime
from ... import __max_supported_opset__, get_ir_version
from .einsum_impl import decompose_einsum_equation, apply_einsum_sequence
from .einsum_ml import predict_transposition_cost


_einsum_cache = {}


def enumerate_cached_einsum():
    """
    Enumerates all cached einsum function.
    """
    global _einsum_cache  # pylint: disable=W0603,W0602
    for k, v in _einsum_cache.items():
        yield k, v


class CachedEinsum:
    """
    Stores all the necessary information to cache the preprocessing
    of a an einsum equation.

    :param equation: numpy equation
    :param runtime: see :func:`einsum
        <mlprodict.testing.einsum.einsum_fct.einsum>`
    :param opset: ONNX opset
    :param optimize: finds the best letter permutation
    :param dtype: dtype
    :param decompose: to decompose Einsum operator or to keep it as is
    :param key: key used to cache this class
    :param strategy: optimization strategy
    :param verbose: displays progress information

    The class creates the following attributes:

    * `equation_` corresponding to the best equivalent equation
    * `graph_`: the corresponding graph returned by function
        :func:`decompose_einsum_equation
        <mlprodict.testing.einsum.einsum_impl.decompose_einsum_equation> `
    * `onnx_`: if a conversion to onnx is used, stores the onnx graph
    * `runtime_`: a function used by `__call__`, calls the runtime
    """

    def __init__(self, equation, runtime='batch_dot', opset=None,
                 optimize=False, dtype=numpy.float64, decompose=True,
                 strategy=None, verbose=None, key=None):
        self.equation = equation
        self.runtime = runtime
        self.opset = opset
        self.optimize = optimize
        self.dtype = dtype
        self.decompose = decompose
        self.strategy = strategy
        self.verbose = verbose
        self.key = key

    def __repr__(self):
        "usual"
        return "%s(%r, %r, %r, %r, %r, %r, %r, key=%r)" % (
            self.__class__.__name__, self.equation, self.runtime,
            self.opset, self.optimize, self.dtype, self.decompose,
            self.strategy, self.key)

    def default_inputs(self, N=None):
        """
        Returns default inputs (reshaped numpy.arange + 0.7i).

        :param N: dimension (all dimension have the same size)

        If *N is None*, N is given a size depending on the number of letters
        to avoid spending too much time on optimization.
        """
        if N is None:
            letters = set(c for c in self.equation
                          if "a" <= c <= "z" or "A" <= c <= "Z")
            nn = math.factorial(len(letters))
            N = max(int(2 ** 11 / nn), 4)
            N = min(N, 15)
        inps = self.equation.split('->')[0].split(',')
        lens = [len(s) for s in inps]
        inputs = [numpy.arange(N ** d).reshape((N,) * d) for d in lens]
        inputs = [(i + 0.7 * ii).astype(self.dtype)
                  for ii, i in enumerate(inputs)]
        return inputs

    def build(self):
        """
        Preprocesses the equation builds whatever is necessary
        to compute the result of the einsum equation.
        """
        if not self.optimize and not hasattr(self, 'equation_'):
            self.equation_ = self.equation
        elif self.strategy is None:
            self.equation_ = self._build_optimize()
        elif self.strategy == 'ml':
            self.equation_ = self._build_optimize_ml()
        else:
            raise ValueError(  # pragma error
                f"Unknown strategy {self.strategy!r}.")
        self.build_runtime()

    def _build_optimize(self):
        # loops over all permutations
        if self.equation.lower() != self.equation:
            raise RuntimeError(  # pragma: no cover
                f"Only lower equation can be optimized, {self.equation!r} is not.")
        letters = list(
            sorted(set(c for c in self.equation if "a" <= c <= "z")))
        possible = list(permutations(letters))
        possible.insert(0, letters)
        if self.verbose:
            from tqdm import tqdm  # pragma: no cover
            subset = tqdm(possible)  # pragma: no cover
        else:
            subset = possible
        best = []
        confs = []
        very_best = None
        inputs = None
        for perm in subset:
            replace = {d: c for c, d in zip(letters, perm)}
            eq = self.equation
            for k, v in replace.items():
                eq = eq.replace(k, v.upper())
            eq = eq.lower()
            inst = CachedEinsum(eq, runtime=self.runtime, opset=self.opset,
                                optimize=False, dtype=self.dtype,
                                decompose=self.decompose)
            inst.build()
            if inputs is None:
                inputs = inst.default_inputs()
                inst(*inputs)
            ts = time.perf_counter()
            for _ in range(0, 10):
                inst(*inputs)
            delta = time.perf_counter() - ts
            confs.append((delta, eq))
            if len(best) < 10:
                best.append((delta, eq))
                best.sort()
            elif delta < best[-1][0]:
                best[-1] = (delta, eq)
                best.sort()
            if self.verbose and (
                    very_best is None or very_best != best[0][0]):
                very_best = best[0][0]
                subset.set_description("%1.2g rtbest=%r" % best[0])
        self.optimized_ = best
        self.timed_permutations_ = confs
        return best[0][1]

    def _build_optimize_ml(self):
        # loops over all permutations
        if self.equation.lower() != self.equation:
            raise RuntimeError(  # pragma: no cover
                f"Only lower equation can be optimized, {self.equation!r} is not.")
        letters = list(
            sorted(set(c for c in self.equation if "a" <= c <= "z")))
        possible = list(permutations(letters))
        possible.insert(0, letters)
        if self.verbose:
            from tqdm import tqdm  # pragma: no cover
            subset = tqdm(possible)  # pragma: no cover
        else:
            subset = possible
        best = []
        confs = []
        very_best = None
        inputs = None
        for perm in subset:
            replace = {d: c for c, d in zip(letters, perm)}
            eq = self.equation
            for k, v in replace.items():
                eq = eq.replace(k, v.upper())
            eq = eq.lower()
            inst = CachedEinsum(eq, runtime=self.runtime, opset=self.opset,
                                optimize=False, dtype=self.dtype,
                                decompose=self.decompose)
            inst.build()
            if inputs is None:
                inputs = inst.default_inputs()
            if hasattr(inst, 'onnx_'):
                onx = inst.onnx_
            else:
                from skl2onnx.common.data_types import FloatTensorType  # delayed
                inits = [
                    ('X%d' % i, FloatTensorType(list(inputs[i].shape)))
                    for i in range(len(inputs))]
                onx = inst.graph_.to_onnx('Y', *inits, opset=self.opset)

            rt = OnnxMicroRuntime(onx)
            dict_inputs = {'X%d' % i: inp for i, inp in enumerate(inputs)}
            out = rt.run(dict_inputs)

            transposes = []
            for node in onx.graph.node:  # pylint: disable=E1101
                if node.op_type == 'Transpose':
                    shape = [(d * 10 if d > 1 else d)
                             for d in out[node.input[0]].shape]
                    transposes.append(
                        [shape, list(node.attribute[0].ints)])

            delta = sum(max(0, predict_transposition_cost(*v))
                        for v in transposes)

            confs.append((delta, eq))
            if len(best) < 10:
                best.append((delta, eq))
                best.sort()
            elif delta < best[-1][0]:
                best[-1] = (delta, eq)
                best.sort()
            if self.verbose and (
                    very_best is None or very_best != best[0][0]):
                very_best = best[0][0]
                subset.set_description("%1.2g mlbest=%r" % best[0])
        self.optimized_ = best
        self.timed_permutations_ = confs
        return best[0][1]

    def build_onnx_einsum(self, input_names):
        """
        Builds an ONNX graph with a single einsum operator.
        """
        opset = (self.opset if self.opset is not None
                 else __max_supported_opset__)
        ir_version = get_ir_version(opset)
        proto_type = guess_proto_dtype(
            numpy.float32 if self.dtype is None else self.dtype)

        model = helper.make_model(
            opset_imports=[helper.make_operatorsetid('', opset)],
            ir_version=ir_version,
            producer_name='mlprodict',
            producer_version='0.0.1',
            graph=helper.make_graph(
                name='einsum',
                inputs=[helper.make_tensor_value_info(n, proto_type, None)
                        for n in input_names],
                outputs=[helper.make_tensor_value_info("Y", proto_type, None)],
                nodes=[
                    helper.make_node(
                        'Einsum', input_names, ["Y"], equation=self.equation_)]))
        return model

    def build_runtime(self):
        """
        Builds the runtime associated to the
        equation `self.equation_`.
        """
        if self.decompose:
            self.graph_ = decompose_einsum_equation(
                self.equation_, strategy='numpy', clean=True)
            if self.runtime == 'batch_dot':
                self.runtime_ = lambda *inputs: apply_einsum_sequence(
                    self.graph_, *inputs)
            elif self.runtime in ('python', 'onnxruntime1'):
                from ...onnxrt import OnnxInference
                n_inputs = len(self.graph_.metadata['lengths']) - 1
                input_names = ['X%d' % i for i in range(n_inputs)]
                self.onnx_names_ = input_names
                onx = self.graph_.to_onnx(
                    'Y', *input_names, opset=self.opset, dtype=self.dtype)
                self.onnx_ = onx
                rt = ('python_compiled'
                      if self.runtime == 'python'
                      else self.runtime)
                self.oinf_ = OnnxInference(
                    self.onnx_, runtime=rt, runtime_options=dict(
                        log_severity_level=3))
                self.runtime_ = lambda *inputs: self.oinf_.run(
                    {i: v for i, v in zip(self.onnx_names_, inputs)})['Y']
            else:
                raise ValueError(  # pragma: no cover
                    f"Unexpected runtime {self.runtime!r}.")
        else:
            if self.runtime in ('python', 'onnxruntime1'):
                from ...onnxrt import OnnxInference
                n_inputs = len(self.equation.split('->')[0].split(','))
                input_names = ['X%d' % i for i in range(n_inputs)]
                self.onnx_ = self.build_onnx_einsum(input_names)
                self.onnx_names_ = input_names
                rt = ('python_compiled'
                      if self.runtime == 'python'
                      else self.runtime)
                self.oinf_ = OnnxInference(
                    self.onnx_, runtime=rt, runtime_options=dict(
                        log_severity_level=3))
                self.runtime_ = lambda *inputs: self.oinf_.run(
                    {i: v for i, v in zip(self.onnx_names_, inputs)})['Y']
            else:
                raise ValueError(  # pragma: no cover
                    f"Unexpected runtime {self.runtime!r}.")

    def __call__(self, *inputs):
        """
        Calls the runtime `self.runtime_`.
        """
        if not hasattr(self, 'runtime_'):
            raise RuntimeError(  # pragma: no cover
                "Method build_runtime was not called.")
        return self.runtime_(*inputs)

    @staticmethod
    def build_einsum(equation, runtime, opset, optimize,
                     dtype, decompose=True, strategy=None,
                     verbose=None, key=None):
        """
        Creates an instance of *CachedEinsum*.
        """
        inst = CachedEinsum(equation, runtime=runtime, opset=opset,
                            optimize=optimize, dtype=dtype,
                            decompose=decompose, strategy=strategy,
                            verbose=verbose, key=key)
        inst.build()
        return inst


def _einsum(equation, dtype, optimize=False, runtime="batch_dot",
            cache=True, opset=None, decompose=True, strategy=None,
            verbose=None):
    global _einsum_cache  # pylint: disable=W0603,W0602
    cached = None
    if cache:
        key = equation, runtime, opset, optimize, dtype, decompose, strategy
        cached = _einsum_cache.get(key, None)
    else:
        key = None
    if cached is None:
        cached = CachedEinsum.build_einsum(
            equation, runtime, opset, optimize,
            dtype, decompose=decompose, strategy=strategy,
            verbose=verbose, key=key)
    else:
        cache = False
    if cache:
        _einsum_cache[key] = cached
    return cached


def optimize_decompose_einsum_equation(
        equation, dtype, optimize=False, runtime="batch_dot",
        cache=True, opset=None, decompose=True, strategy=None,
        verbose=None):
    """
    Proposes a new implementation of :epkg:`numpy:einsum`.
    It does not allow expresion using `...` and expects
    a right member.

    :param equation: einsum equation
    :param optimize: permutes all letters to find the best
        permutation
    :param runtime: runtime used to compute the results once the
        computation graph is produced (see below)
    :param cache: if True, the function stores the preprocessing
        done for a specific equation, the second call with the same
        equation is much faster
    :param opset: ONNX opset to use for some runtimes
    :param decompose: by default, the function decomposes
        the equation into more simple operators but it can keep
        the original ONNX einsum operator.
    :param strategy: optimisation strategy (see below)
    :param verbose: display progress if optimize is True
    :return: einsum result

    The available runtimes are:

    * `batch_dot`: the runtime is @see fn apply_einsum_sequence,
    * `python`: one ONNX graph executed with a python runtime,
    * `onnxruntime1`: one ONNX graph executed with :epkg:`onnxruntime`.

    The optimisation strategy can be:

    * `None`: the same runtime is used to find the best permutation of letters
    * `'ml'`: a machine learned model is used to predict the
        best permutation of letters, this model comes from
        notebook :ref:`onnxoperatorcostrst`.

    The function works in two steps:

    * first step analyses the equation to produce a computation graph,
      this graph can also be converted into ONNX,
    * second step runs the graph whatever the graph is.

    The function returns an object of type @see cl CachedEinsum
    which has the following members after optimization:

    * `equation_` corresponding to the best equivalent equation
    * `graph_`: the corresponding graph returned by function
        :func:`decompose_einsum_equation
        <mlprodict.testing.einsum.einsum_impl.decompose_einsum_equation> `
    * `onnx_`: if a conversion to onnx is used, stores the onnx graph
    * `runtime_`: a function used by `__call__`, calls the runtime
    * `oinf_`: an object of type @see cl OnnxInference
    * `timed_permutations_`: memorizes the results of the optimization

    .. runpython::
        :showcode:

        import numpy
        from mlprodict.testing.einsum import optimize_decompose_einsum_equation

        seq_opt = optimize_decompose_einsum_equation(
            "bsnh,btnh->bnts", numpy.float64, strategy='ml', verbose=1,
            runtime="python", optimize=True)

        print("best equation:", seq_opt.equation_)

    """
    res = _einsum(equation, dtype, optimize=optimize, runtime=runtime,
                  cache=cache, opset=opset, decompose=decompose,
                  strategy=strategy, verbose=verbose)
    return res


def einsum(equation, *inputs, optimize=False, runtime="batch_dot",
           cache=True, opset=None, decompose=True,
           strategy=None, verbose=None):
    """
    Proposes a new implementation of :epkg:`numpy:einsum`.
    It does not allow expresion using `...` and expects
    a right member.

    :param equation: einsum equation
    :param inputs: inputs
    :param optimize: permutes all letters to find the best
        permutation
    :param runtime: runtime used to compute the results once the
        computation graph is produced (see below)
    :param cache: if True, the function stores the preprocessing
        done for a specific equation, the second call with the same
        equation is much faster
    :param opset: ONNX opset to use for some runtimes
    :param decompose: by default, the function decomposes
        the equation into more simple operators but it can keep
        the original ONNX einsum operator.
    :param strategy: optimisation strategy (see below)
    :param verbose: display progress if optimize is True
    :return: einsum result

    The available runtimes are:

    * `batch_dot`: the runtime is @see fn apply_einsum_sequence,
    * `python`: one ONNX graph executed with a python runtime,
    * `onnxruntime1`: one ONNX graph executed with :epkg:`onnxruntime`.

    The optimisation strategy can be:

    * `None`: the same runtime is used to find the best permutation of letters
    * `'ml'`: a machine learned model is used to predict the
        best permutation of letters, this model comes from
        notebook :ref:`onnxoperatorcostrst`.

    The function works in two steps:

    * first step analyses the equation to produce a computation graph,
      this graph can also be converted into ONNX,
    * second step runs the graph whatever the graph is.

    Further details are available in the documentation of function
    @see fn optimize_decompose_einsum_equation.
    The function works the same way as :epkg:`numpy:einsum`:

    .. runpython::
        :showcode:

        import numpy
        from mlprodict.testing.einsum import einsum

        equation = "abc,cd->abd"

        m1 = numpy.random.randn(2, 2, 2)
        m2 = numpy.random.randn(2, 2)

        np = numpy.einsum(equation, m1, m2)
        print('numpy.einsum')
        print(np)

        print('mlprodict.testing.einsum')
        mp = einsum(equation, m1, m2)
        print(mp)

    In some case, the einsum implementation can be optimized by looping
    on possible permutation:

    .. runpython::
        :showcode:
        :process:

        import timeit
        import numpy
        from mlprodict.testing.einsum import einsum
        from mlprodict.testing.einsum.einsum_fct import enumerate_cached_einsum

        equation = "cab,cd->ad"

        m1 = numpy.random.randn(20, 20, 20)
        m2 = numpy.random.randn(20, 20)

        print('numpy.einsum',
              timeit.timeit('numpy.einsum(equation, m1, m2)',
                            number=200,
                            globals=globals()))

        einsum(equation, m1, m2)
        print('einsum',
              timeit.timeit('einsum(equation, m1, m2)',
                            number=200,
                            globals=globals()))

        einsum(equation, m1, m2, runtime='python')
        print('einsum-python',
              timeit.timeit('einsum(equation, m1, m2, runtime="python")',
                            number=200,
                            globals=globals()))

        einsum(equation, m1, m2, runtime='onnxruntime1')
        print('einsum-onnxruntime1',
              timeit.timeit('einsum(equation, m1, m2, runtime="onnxruntime1")',
                            number=200,
                            globals=globals()))

        einsum(equation, m1, m2, runtime='onnxruntime1', optimize=True, verbose=1)
        print('einsum-onnxruntime1',
              timeit.timeit('einsum(equation, m1, m2, runtime="onnxruntime1", optimize=True)',
                            number=200,
                            globals=globals()))

        print("list of cached einsum equations")
        for k, v in enumerate_cached_einsum():
            print(k, v.equation, v.equation_)

    The last example shows the time taken by every function:

    .. runpython::
        :showcode:
        :process:

        import os
        from pyquickhelper.pycode.profiling import profile
        import numpy
        from mlprodict.testing.einsum import einsum
        from mlprodict.testing.einsum.einsum_fct import enumerate_cached_einsum
        from mlprodict import __file__ as path

        root = os.path.dirname(path)

        equation = "cab,cd->ad"

        m1 = numpy.random.randn(200, 20, 20)
        m2 = numpy.random.randn(200, 20)

        def clean(txt):
            txt = txt.replace(root, "mlprodict")
            return "\\n".join(txt.split("\\n")[:30])

        def fct1():
            for i in range(100):
                einsum(equation, m1, m2, cache=False)

        print("Profile cache with default runtime.")
        res = profile(fct1)
        print(root)
        print(clean(res[1]))

        def fct2():
            for i in range(100):
                einsum(equation, m1, m2, cache=False, runtime='python')

        print("Profile cache with runtime='python'.")
        res = profile(fct2)
        print(root)
        print(clean(res[1]))


        def fct3():
            for i in range(100):
                einsum(equation, m1, m2, cache=True)

        einsum(equation, m1, m2, cache=True)
        print("Profile execution with default runtime.")
        res = profile(fct3)
        print(root)
        print(clean(res[1]))



        def fct4():
            for i in range(100):
                einsum(equation, m1, m2, cache=True, runtime='python')

        einsum(equation, m1, m2, cache=True, runtime='python')
        print("Profile execution with runtime='python'.")
        res = profile(fct4)
        print(root)
        print(clean(res[1]))


        def fct5():
            for i in range(100):
                einsum(equation, m1, m2, cache=True, runtime='onnxruntime1')

        einsum(equation, m1, m2, cache=True, runtime='onnxruntime1')
        print("Profile execution with runtime='onnxruntime1'.")
        res = profile(fct5)
        print(root)
        print(clean(res[1]))
    """
    if len(inputs) == 0:
        raise ValueError("No inputs found.")  # pragma: no cover
    dtypes = set(i.dtype for i in inputs)
    if len(dtypes) != 1:
        raise ValueError(  # pragma: no cover
            "All inputs do not have the same type (%r), "
            "all of them should be cast before called einsum."
            "" % dtypes)
    cached = optimize_decompose_einsum_equation(
        equation, inputs[0].dtype, optimize=optimize,
        runtime=runtime, cache=cache, opset=opset,
        decompose=decompose, strategy=strategy, verbose=verbose)
    return cached(*inputs)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class _CommonQuantizeLinear(OpRun):

    def __init__(self, onnx_node, desc=None,
                 expected_attributes=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=expected_attributes,
                       **options)

    def common_run(self, x, y_scale, zero_point=None, axis=1):  # pylint: disable=W0221
        if len(y_scale.shape) > 1:
            raise RuntimeError(  # pragma: no cover
                "Input 2 must be a vector or a number.")
        if len(y_scale.shape) > 0 and y_scale.size == 1:
            y_scale = y_scale[0]
        if len(y_scale.shape) > 0:
            new_shape = [1 for s in x.shape]
            new_shape[axis] = len(y_scale)
            x = x / y_scale.reshape(new_shape)
        else:
            x = x / y_scale
        if zero_point is not None:
            dtype = zero_point.dtype
            if len(y_scale.shape) > 0:
                x += zero_point.reshape(new_shape)
            else:
                x += zero_point
            numpy.around(x, 1, out=x)
            if dtype == numpy.uint8:
                numpy.clip(x, 0, 255, out=x)
            elif dtype == numpy.int8:
                numpy.clip(x, -128, 127, out=x)
            else:
                raise RuntimeError(  # pragma no cover
                    f"Unexpected dtype for input 2 {dtype}.")
            return (x.astype(dtype), )

        dtype = numpy.uint8
        numpy.around(x, 1, out=x)
        numpy.clip(x, 0, 255, out=x)
        return (x.astype(dtype), )


class QuantizeLinear(_CommonQuantizeLinear):

    atts = {'axis': 1}
    python_inputs = ['*inputs']

    def __init__(self, onnx_node, desc=None, **options):
        _CommonQuantizeLinear.__init__(
            self, onnx_node, desc=desc,
            expected_attributes=QuantizeLinear.atts,
            **options)

    def _run(self, *args, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        # args: x, y_scale, zero_point
        return self.common_run(*args, axis=self.axis)


class DynamicQuantizeLinear(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       **options)
        self.dtype = numpy.uint8

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        # args: x, y_scale, zero_point
        qmin, qmax = 0, 255
        minx = numpy.min(x)
        y_scale = (numpy.max(x) - minx) / (qmax - qmin)
        intermediate_zero_point = qmin - minx / y_scale
        y_zero_point = numpy.round(
            numpy.clip(intermediate_zero_point, qmin, qmax)).astype(self.dtype)
        y = numpy.clip(numpy.round(x / y_scale) + y_zero_point, qmin, qmax)
        return (y.astype(self.dtype),
                y_scale.astype(x.dtype),
                y_zero_point.astype(self.dtype))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from ._new_ops import OperatorSchema


class FusedMatMul(OpRun):

    atts = {'alpha': 1., 'transA': 0, 'transB': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=FusedMatMul.atts,
                       **options)
        if self.transA:
            _meth = (FusedMatMul._fmatmul11 if self.transB
                     else FusedMatMul._fmatmul10)
        else:
            _meth = (FusedMatMul._fmatmul01 if self.transB
                     else FusedMatMul._fmatmul00)
        self._meth = lambda a, b: _meth(a, b, self.alpha)

    def _find_custom_operator_schema(self, op_name):
        if op_name == "FusedMatMul":
            return FusedMatMulSchema()
        raise RuntimeError(  # pragma: no cover
            f"Unable to find a schema for operator '{op_name}'.")

    @staticmethod
    def _fmatmul00(a, b, alpha):
        return numpy.matmul(a, b) * alpha

    @staticmethod
    def _fmatmul01(a, b, alpha):
        return numpy.matmul(a, b.T) * alpha

    @staticmethod
    def _fmatmul10(a, b, alpha):
        return numpy.matmul(a.T, b) * alpha

    @staticmethod
    def _fmatmul11(a, b, alpha):
        return numpy.matmul(a.T, b.T) * alpha

    def _run(self, a, b, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (self._meth(a, b), )


class FusedMatMulSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl FusedMatMul.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'FusedMatMul')
        self.attributes = FusedMatMul.atts



"""
@file
@brief Optimisation of :epkg:`ONNX` graphs.
"""
import logging
from onnx import FunctionProto, GraphProto
from onnx.helper import make_graph, make_function
from ._onnx_optimisation_common import (  # pylint: disable=E0611
    _apply_optimisation_on_graph, _apply_remove_node_fct_node)


logger = logging.getLogger('onnx:optim')


def _process_node(node, data, edges, paths, prefix="", sep="::", path=None):
    node_name = prefix + node.name
    data[node_name, 1] = node
    path = [] if path is None else path.copy()
    paths[node_name, 1] = path
    path = path.copy()
    path.append(node_name)
    for inp in node.input:
        data[inp, 0] = node
        edges[(inp, 0), (node_name, 1)] = node
        paths[inp, 0] = path
        if '::' in node_name:
            # We need to link an input to the parent node
            # if the node is part of subgraph.
            # path_r = paths[inp, 0]
            if len(path) <= 1:
                raise RuntimeError(  # pragma: no cover
                    f"Unexpected path {path!r}.")
            edges[(inp, 0), (path[-2], 1)] = node

    for out in node.output:
        data[out, 0] = node
        paths[out, 0] = node_name
        edges[(node_name, 1), (out, 0)] = node
    if len(node.attribute) > 0:
        for att in node.attribute:
            if not hasattr(att, 'g'):
                continue
            if not isinstance(att.g, GraphProto):
                continue
            for no in att.g.node:
                _process_node(no, data, edges, paths,
                              prefix=node_name + sep, path=path)


def onnx_remove_node_unused(onnx_model, recursive=True, debug_info=None, **options):
    """
    Removes unused nodes of the graph. An unused node
    is not involved in the output computation.

    :param onnx_model: onnx model
    :param recursive: looks into subgraphs
    :param debug_info: debug information (private)
    :param options: unused
    :return: new onnx _model
    """
    if debug_info is None:
        debug_info = [str(type(onnx_model)).rsplit(
            '.', maxsplit=1)[-1].strip("'>")]
    else:
        debug_info = (debug_info +
                      [str(type(onnx_model)).rsplit('.', maxsplit=1)[-1].strip("'>")])

    if hasattr(onnx_model, 'graph'):
        return _apply_optimisation_on_graph(
            onnx_remove_node_unused, onnx_model,
            recursive=recursive, debug_info=debug_info,
            **options)

    graph = onnx_model
    logger.debug("onnx_remove_node_unused:begin with %d nodes.",
                 len(graph.node))
    is_function = isinstance(graph, FunctionProto)
    data = {}
    valid = {}
    edges = {}
    paths = {}

    if not is_function:
        for init in graph.initializer:
            data[init.name, 0] = init

    for node in graph.node:
        _process_node(node, data, edges, paths)

    for out in graph.output:
        valid[out if is_function else out.name, 0] = True

    modif = 1
    while modif > 0:
        modif = 0
        for e1, e2 in edges:  # pylint: disable=E1141
            if valid.get(e2, False) and not valid.get(e1, False):
                valid[e1] = True
                modif += 1

    new_nodes = [n for n in graph.node if (n.name, 1) in valid]
    if not is_function:
        new_inits = [n for n in graph.initializer if (n.name, 0) in valid]

    if recursive:
        # Handles subgraphs.
        for i in range(len(new_nodes)):  # pylint: disable=C0200
            node = new_nodes[i]
            if node is None or not (node.attribute):  # pylint: disable=C0325
                continue
            new_nodes[i] = _apply_remove_node_fct_node(
                onnx_remove_node_unused,
                node, recursive=True, debug_info=debug_info + [node.name])

    # Finally create the new graph.
    nodes = list(filter(lambda n: n is not None, new_nodes))
    if is_function:
        logger.debug("onnx_remove_node_unused:end function with %d nodes.",
                     len(nodes))
        return make_function(
            onnx_model.domain, onnx_model.name,
            onnx_model.input, onnx_model.output, nodes,
            opset_imports=onnx_model.opset_import,
            attributes=onnx_model.attribute,
            doc_string=onnx_model.doc_string)
    graph = make_graph(nodes, onnx_model.name,
                       onnx_model.input, onnx_model.output,
                       new_inits)
    graph.value_info.extend(onnx_model.value_info)  # pylint: disable=E1101
    logger.debug("onnx_remove_node_unused:end graph with %d nodes.",
                 len(nodes))
    return graph



"""
@file
@brief Intermediate class between :epkg:`numpy` and :epkg:`onnx`.

.. versionadded:: 0.6
"""
import logging
import numpy
from onnx.helper import make_tensor
from ..onnx_tools.onnx2py_helper import guess_proto_dtype
from .xop_variable import Variable
from .xop import loadop, OnnxOperatorItem, OnnxOperatorTuple
from .xop_variable import guess_numpy_type

logger = logging.getLogger('xop')


try:
    numpy_bool = numpy.bool_
except AttributeError:  # pragma: no cover
    numpy_bool = bool
try:
    numpy_str = numpy.str_
except AttributeError:  # pragma: no cover
    numpy_str = str


class OnnxVar:
    """
    Variables used into :epkg:`onnx` computation.

    :param inputs: variable name or object
    :param op: :epkg:`ONNX` operator
    :param select_output: if multiple output are returned by
        ONNX operator *op*, it takes only one specifed by this
        argument
    :param dtype: specifies the type of the variable
        held by this class (*op* is None) in that case
    :param kwargs: addition argument to give operator *op*

    .. versionadded:: 0.6
    """
    __array_ufunc__ = None

    def __init__(self, *inputs, op=None, select_output=None,
                 dtype=None, **kwargs):
        logger.debug('OnnxVar(%d in, dtype=%r, op=%r, select_output=%r)',
                     len(inputs), dtype, op, select_output)
        self.inputs = inputs
        self.select_output = select_output
        self.onnx_op = op
        self.alg_ = None
        self.onnx_op_kwargs = kwargs
        if dtype is not None and (op is not None or len(inputs) != 1):
            raise RuntimeError(  # pragma: no cover
                "dtype can only be used if op is None or len(inputs) == 1.")
        for i, inp in enumerate(self.inputs):
            if isinstance(inp, type):
                raise TypeError(  # pragma: no cover
                    "Unexpected type for input %d - %r." % (i, inp))
            if not isinstance(inp, numpy.ndarray):
                continue
            if (inp.size > 0 and
                    isinstance(inp.ravel()[0], (numpy.ndarray, OnnxVar))):
                raise TypeError(  # pragma: no cover
                    "Unexpected type for input %d: %r, %r, "
                    "op=%r" % (i, type(inp), inp.ravel()[0], op))
        self.dtype = self._guess_dtype(dtype, from_init=True)

    def _guess_dtype(self, dtype, from_init=False):
        "Guesses dtype when not specified."
        if dtype is not None:
            return dtype
        dtypes = []
        for i, inp in enumerate(self.inputs):
            if isinstance(inp, str):
                return None
            if isinstance(inp, numpy.ndarray):
                dtypes.append(inp.dtype)
            elif isinstance(inp, Variable):
                dtypes.append(inp.dtype)
            elif isinstance(inp, OnnxVar):
                dtypes.append(inp.dtype)
            elif isinstance(inp, MultiOnnxVar):
                dtypes.append(inp._guess_dtype(dtype))
            elif isinstance(inp, (numpy.float32, numpy.float64,
                                  numpy.int32, numpy.int64)):
                dtypes.append(inp.dtype)
            elif isinstance(inp, numpy_str):
                dtypes.append(numpy_str)
            elif isinstance(inp, numpy_bool):
                dtypes.append(numpy_bool)
            elif isinstance(inp, int):
                dtypes.append(numpy.int64)  # pragma: no cover
            elif isinstance(inp, float):
                dtypes.append(numpy.float64)
            elif hasattr(inp, 'fit'):
                # scikit-learn model
                continue
            elif hasattr(inp, '_guess_dtype'):
                dtypes.append(inp._guess_dtype(dtype))
            else:
                try:
                    dtype = guess_numpy_type(inp)
                except NotImplementedError as e:
                    raise TypeError(  # pragma: no cover
                        "Unexpected type for input %i type=%r." % (
                            i, type(inp))) from e
                dtypes.append(dtype)
        dtypes = [_ for _ in dtypes if _ is not None]
        unique = set(dtypes)
        if len(unique) != 1:
            return None
        return dtypes[0]

    def __repr__(self):
        "usual"
        args = []
        for inp in self.inputs:
            args.append(repr(inp))
        if self.onnx_op is not None:
            if isinstance(self.onnx_op, str):
                args.append(f"op={self.onnx_op!r}")
            else:
                args.append(f"op={self.onnx_op.__name__}")
        if self.select_output is not None:
            args.append(f"select_output={self.select_output!r}")
        if self.dtype is not None and self.dtype != self._guess_dtype(None):
            args.append(f"dtype={self.dtype!r}")
        for k, v in sorted(self.onnx_op_kwargs.items()):
            args.append(f"{k}={v!r}")
        res = f"{self.__class__.__name__}({', '.join(args)})"
        return res

    def set_onnx_name(self, name_type):
        """
        Forces this variable to get this name during

        :param name_type: a tuple *(name, type)*
        """
        self.onnx_input_type_ = name_type

    def to_algebra(self, op_version=None):
        """
        Converts the variable into an operator.
        """
        if self.alg_ is not None:
            return self.alg_

        if self.onnx_op is None:
            logger.debug('OnnxVar.to_algebra:1(op_version=%r)', op_version)
            if len(self.inputs) != 1:
                raise RuntimeError(  # pragma: no cover
                    "Unexpected number of inputs, 1 expected, "
                    "got {} instead.".format(self.inputs))
            if self.dtype is None or hasattr(self.inputs[0], 'onnx_name'):
                self.alg_ = Variable.from_skl2onnx(self.inputs[0])
            elif isinstance(self.inputs[0], Variable):
                self.alg_ = self.inputs[0]
            else:
                self.alg_ = Variable(self.inputs[0], self.dtype)
        else:
            logger.debug('OnnxVar.to_algebra:2(op_version=%r) - onnx_op=%r',
                         op_version, self.onnx_op)
            if isinstance(self.onnx_op, str):
                var = self._custom_op(*self.inputs, op_version=op_version,
                                      **self.onnx_op_kwargs)
                alg = var.to_algebra(op_version=op_version)
                if not hasattr(self, 'alg_'):
                    raise RuntimeError(  # pragma: no cover
                        "Missing attribute 'alg_'.")
                self.alg_ = alg
                return alg

            new_inputs = []
            for inp in self.inputs:
                if hasattr(inp, 'fit'):
                    # scikit-learn model
                    new_inputs.append(inp)
                elif isinstance(inp, (
                        int, float, str, numpy.ndarray, numpy.int32,
                        numpy.int64, numpy.float32, numpy.float64,
                        numpy_bool, numpy_str, numpy.int8, numpy.uint8,
                        numpy.int16, numpy.uint16, numpy.uint32,
                        numpy.uint64)):
                    if (inp.size > 0 and
                            isinstance(
                                inp.ravel()[0],  # pylint: disable=E1101
                                (numpy.ndarray, OnnxVar))):
                        raise TypeError(  # pragma: no cover
                            "Unexpected type for an input %r, %r."
                            "" % (type(inp), inp.ravel()[0]))  # pylint: disable=E1101
                    new_inputs.append(inp)
                else:
                    new_inputs.append(
                        inp.to_algebra(op_version=op_version))

            res = self.onnx_op(*new_inputs, op_version=op_version,
                               **self.onnx_op_kwargs)
            if self.select_output is None:
                self.alg_ = res
            else:
                self.alg_ = res[self.select_output]
        return self.alg_

    def _custom_op(self, *args, op_version=None, runtime=None, **kwargs):
        """
        This could be handled before a call to this method
        but this method can change the conversion of an non-existing
        operator depending on the given opset.
        """
        if self.onnx_op == 'filter':
            return self._custom_op_filter(*args, op_version=op_version,
                                          runtime=runtime, **kwargs)
        raise NotImplementedError(  # pragma: no cover
            f"Unexpected custom operator {self.onnx_op!r}.")

    def _custom_op_filter(self, *args, op_version=None, runtime=None, **kwargs):
        """
        This could be handled before a call to this method
        but this method can change the conversion of an non-existing
        operator depending on the given opset.
        """
        OnnxSqueeze, OnnxTopK, OnnxGather, OnnxReduceSum = loadop(
            'Squeeze', 'TopK', 'Gather', 'ReduceSum')
        if len(args) != 2:
            raise RuntimeError(  # pragma: no cover
                f"Custom op 'filter' expects two inputs not {len(args)!r}.")
        if len(kwargs) != 0:
            raise RuntimeError(  # pragma: no cover
                f"Custom op 'filter' expects no arguments but got {kwargs!r}.")
        mat, index = args
        cast = OnnxVar(index.astype(numpy.int64), op=OnnxSqueeze)
        n1 = OnnxVar(cast, op=OnnxReduceSum, keepdims=1)
        indices = OnnxVar(cast, n1, op=OnnxTopK, select_output=1)
        return OnnxVar(mat, indices, op=OnnxGather)

    @property
    def T(self):
        "Transpose."
        OnnxTranspose = loadop('Transpose')
        return OnnxVar(self, op=OnnxTranspose)

    def astype(self, dtype):
        "Cast"
        OnnxCast = loadop('Cast')
        return OnnxVar(self, op=OnnxCast, to=guess_proto_dtype(dtype))

    @property
    def shape(self):
        "Shape"
        OnnxShape = loadop('Shape')
        return OnnxVar(self, op=OnnxShape)

    @property
    def size(self):
        "Size"
        OnnxSize = loadop('Size')
        return OnnxVar(self, op=OnnxSize)

    def reshape(self, shape):
        "Reshape"
        OnnxReshape = loadop('Reshape')
        if isinstance(shape, (tuple, list)):
            shape = numpy.array(shape, dtype=numpy.int64)
        return OnnxVar(self, shape, op=OnnxReshape)

    def _make_array(self, y):
        """Converts *y* into an array if not."""
        if isinstance(y, (numpy.ndarray, OnnxVar)):
            return y
        if hasattr(y, 'dtype'):
            return numpy.full((1, ), y, dtype=y.dtype)
        if isinstance(y, str):
            return numpy.array([y])
        if isinstance(y, float):
            return numpy.array([y], dtype=numpy.float32)
        if isinstance(y, int):
            return numpy.array([y], dtype=numpy.int64)
        return y

    def __add__(self, y):
        "Addition."
        y = self._make_array(y)
        OnnxAdd = loadop('Add')
        return OnnxVar(self, y, op=OnnxAdd)

    def __radd__(self, y):
        "Right Addition."
        y = self._make_array(y)
        OnnxIdentity, OnnxAdd = loadop('Identity', 'Add')
        return OnnxVar(OnnxVar(y, op=OnnxIdentity), self, op=OnnxAdd)

    def __sub__(self, y):
        "Subtraction."
        y = self._make_array(y)
        OnnxSub = loadop('Sub')
        return OnnxVar(self, y, op=OnnxSub)

    def __rsub__(self, y):
        "Right subtraction."
        y = self._make_array(y)
        OnnxIdentity, OnnxSub = loadop('Identity', 'Sub')
        return OnnxVar(OnnxVar(y, op=OnnxIdentity), self, op=OnnxSub)

    def __mul__(self, y):
        "Multiplication."
        y = self._make_array(y)
        OnnxMul = loadop('Mul')
        return OnnxVar(self, y, op=OnnxMul)

    def __rmul__(self, y):
        "Right multiplication."
        y = self._make_array(y)
        OnnxIdentity = loadop('Identity')
        return OnnxVar(y, op=OnnxIdentity) * self

    def __pow__(self, y):
        "Power."
        y = self._make_array(y)
        OnnxPow = loadop('Pow')
        return OnnxVar(self, y, op=OnnxPow)

    def __mod__(self, y):
        "Modulo."
        y = self._make_array(y)
        OnnxMod = loadop('Mod')
        return OnnxVar(self, y, op=OnnxMod)

    def __matmul__(self, y):
        "Matrix multiplication."
        y = self._make_array(y)
        OnnxMatMul = loadop('MatMul')
        return OnnxVar(self, y, op=OnnxMatMul)

    def __truediv__(self, y):
        "Division, no difference between `/` and `//`."
        y = self._make_array(y)
        OnnxDiv = loadop('Div')
        return OnnxVar(self, y, op=OnnxDiv)

    def __rtruediv__(self, y):
        "Division, no difference between `/` and `//`."
        y = self._make_array(y)
        OnnxIdentity, OnnxDiv = loadop('Identity', 'Div')
        return OnnxVar(OnnxVar(y, op=OnnxIdentity), self, op=OnnxDiv)

    def __floordiv__(self, y):
        "Division, no difference between `/` and `//`."
        y = self._make_array(y)
        OnnxDiv = loadop('Div')
        return OnnxVar(self, y, op=OnnxDiv)

    def __eq__(self, y):
        "Equality."
        y = self._make_array(y)
        OnnxEqual = loadop('Equal')
        return OnnxVar(self, y, op=OnnxEqual)

    def __ne__(self, y):
        "Difference."
        y = self._make_array(y)
        OnnxEqual, OnnxNot = loadop('Equal', 'Not')
        return OnnxVar(OnnxVar(self, y, op=OnnxEqual), op=OnnxNot)

    def __ge__(self, y):
        "Greater or Equal."
        y = self._make_array(y)
        OnnxGreaterOrEqual = loadop('GreaterOrEqual')
        return OnnxVar(self, y, op=OnnxGreaterOrEqual)

    def __gt__(self, y):
        "Greater."
        y = self._make_array(y)
        OnnxGreater = loadop('Greater')
        return OnnxVar(self, y, op=OnnxGreater)

    def __invert__(self):
        "not."
        OnnxNot = loadop('Not')
        return OnnxVar(self, op=OnnxNot)

    def __le__(self, y):
        "Less or Equal."
        y = self._make_array(y)
        OnnxLessOrEqual = loadop('LessOrEqual')
        return OnnxVar(self, y, op=OnnxLessOrEqual)

    def __lt__(self, y):
        "Less."
        y = self._make_array(y)
        OnnxLess = loadop('Less')
        return OnnxVar(self, y, op=OnnxLess)

    def __and__(self, y):
        "And."
        y = self._make_array(y)
        OnnxAnd = loadop('And')
        return OnnxVar(self, y, op=OnnxAnd)

    def __or__(self, y):
        "And."
        y = self._make_array(y)
        OnnxOr = loadop('Or')
        return OnnxVar(self, y, op=OnnxOr)

    def not_(self):
        "Not."
        OnnxNot = loadop('Not')
        return OnnxVar(self, op=OnnxNot)

    def __neg__(self):
        "Neg."
        OnnxNeg = loadop('Neg')
        return OnnxVar(self, op=OnnxNeg)

    def __getitem__(self, index):
        """
        Deals with multiple scenarios.

        * *index* is an integer or a slice, a tuple of integers and slices,
          example: `[0, 1]`, `[:5, :6]`, `[::2]` (**scenario 1**)
        * *index* is an *ONNX* object (more precisely an instance of
          @see cl OnnxVar), then the method assumes it is an array of
          boolean to select a subset of the tensor along the first axis,
          example: `mat[mat == 0]` (**scenario 2**)
        """
        if isinstance(index, OnnxVar):
            # scenario 2
            return OnnxVar(self, index, op='filter')

        if isinstance(index, int):
            # Use Gather instead.
            OnnxGather = loadop('Gather')
            return OnnxVar(
                self, numpy.array(index, dtype=numpy.int64),
                axis=0, op=OnnxGather)

        if not isinstance(index, tuple):
            index = (index, )

        # only one integer?
        ni = None
        ax = None
        for i, a in enumerate(index):
            if isinstance(a, int):
                if ni is None:
                    ni = i
                    ax = a
                else:
                    ax = None
                    ni = None
                    break
            if (isinstance(a, slice) and a.start is None and
                    a.stop is None and a.step is None):
                continue
            ax = None
            ni = None
            break
        if ni is not None and ax is not None:
            # Use Gather instead.
            OnnxGather = loadop('Gather')
            return OnnxVar(
                self, numpy.array(ni, dtype=numpy.int64),
                axis=ax, op=OnnxGather)

        # scenario 1
        starts = []
        ends = []
        axes = []
        steps = []
        axis_squeeze = []
        needs_shape = []
        for i, ind in enumerate(index):
            if isinstance(ind, int):
                starts.append(ind)
                ends.append(ind + 1)
                axes.append(i)
                steps.append(1)
                axis_squeeze.append(i)
                continue
            if isinstance(ind, slice):
                if ind.start is None and ind.stop is None and ind.step is None:
                    continue
                start = 0 if ind.start is None else ind.start
                end = (None, i) if ind.stop is None else ind.stop
                step = 1 if ind.step is None else ind.step
                starts.append(start)
                ends.append(end)
                axes.append(i)
                steps.append(step)
                if isinstance(end, tuple):
                    needs_shape.append(len(ends) - 1)
                elif isinstance(end, OnnxVar):
                    needs_shape.append(end)
                continue
            raise NotImplementedError(  # pragma: no cover
                f"Not implemented for type {type(ind)!r}.")

        if max(steps) == min(steps) == 1:
            steps = None
        else:
            steps = numpy.array(steps, dtype=numpy.int64)

        starts = numpy.array(starts, dtype=numpy.int64)
        axes = numpy.array(axes, dtype=numpy.int64)

        OnnxGather, OnnxSlice, OnnxSqueeze, OnnxConcat = loadop(
            'Gather', 'Slice', 'Squeeze', 'Concat')
        if len(needs_shape) > 0:
            shape = self.shape
            conc = []
            for e in ends:
                if isinstance(e, tuple):
                    conc.append(
                        OnnxVar(shape, numpy.array([e[1]], numpy.int64),
                                op=OnnxGather))
                elif isinstance(e, OnnxVar):
                    conc.append(
                        e.reshape(numpy.array([-1], dtype=numpy.int64)))
                else:
                    conc.append(numpy.array([e], dtype=numpy.int64))
            if len(conc) > 1:
                ends = OnnxVar(*conc, op=OnnxConcat, axis=0)
            else:
                ends = conc[0]
        else:
            ends = numpy.array(ends, dtype=numpy.int64)

        if steps is None:
            sliced = OnnxVar(self, starts, ends, axes, op=OnnxSlice)
        else:
            sliced = OnnxVar(self, starts, ends, axes, steps, op=OnnxSlice)
        if len(axis_squeeze) > 0:
            return OnnxVar(
                sliced, numpy.array(axis_squeeze, dtype=numpy.int64),
                op=OnnxSqueeze)
        return sliced

    def __setitem__(self, index, value):
        """
        Only supports vectors (1D tensor).

        * *index* is an integer or a slice, a tuple of integers and slices,
          example: `[0]`, `[:5]`, `[::2]` (**scenario 1**)
        * *index* is an *ONNX* object (more precisely an instance of
          @see cl OnnxVar), then the method assumes it is an array of
          boolean to select a subset of the tensor along the first axis,
          example: `mat[mat == 0]` (**scenario 2**)
        This processing is applied before the operator it contains.
        A copy should be made (Identity node or copy method).
        """
        OnnxIdentity = loadop('Identity')
        if self.onnx_op is not None and self.onnx_op is not OnnxIdentity:
            raise RuntimeError(  # pragma: no cover
                "A copy should be made before setting new values on a matrix. "
                "Method copy() would do that.")

        if isinstance(index, OnnxVar):
            # scenario 2, example: cp[x < 0] = -1
            return self._setitem2i_(index, value)
        elif not isinstance(index, tuple):
            index = (index, )

        for i in index:
            if isinstance(i, OnnxVar):
                raise NotImplementedError(  # pragma: no cover
                    "Unable to handle case such as cp[0, x < 0] = -1.")

        # scenario 1
        if len(index) == 1:
            return self._setitem1i_(index[0], value)
        raise NotImplementedError(  # pragma: no cover
            f"Indices in {len(index)} dimensions are not implemented yet.")

    def _setitem1i_(self, index, value):
        sl = None
        if isinstance(index, slice):
            start = 0 if index.start is None else index.start
            stop = index.stop
            step = index.step
            sl = [start, stop, step]
        elif isinstance(index, int):
            sl = [index, index + 1, 1]
        else:
            raise NotImplementedError(  # pragma: no cover
                f"Unable to assign new values due to unexpected type {type(index)!r}.")

        if sl[1] is None and isinstance(value, numpy.ndarray):
            sl[1] = sl[0] + value.size
        OnnxConstantOfShape, OnnxScatterElements = loadop(
            'ConstantOfShape', 'ScatterElements')
        if sl[1] is None:
            if sl[2] is not None and sl[2] != 1:
                raise NotImplementedError(  # pragma: no cover
                    "If the length is not known, step must be 1 not %d." % sl[2])
            value = make_tensor(
                "value", guess_proto_dtype(value.dtype), (1, ), [value])  # pylint: disable=E1101
            inp = self.inputs[0]
            if not isinstance(inp, OnnxVar):
                raise RuntimeError(  # pragma: no cover
                    f"Input must be an instance of OnnxVar not {type(inp)!r}.")
            cst = OnnxVar(inp.shape, op=OnnxConstantOfShape, value=value)
            ext = inp[:sl[0]]
            indices = numpy.arange(0, sl[0]).astype(numpy.int64)
            add_step = OnnxVar(cst, indices, ext,
                               op=OnnxScatterElements, axis=0)
        else:
            indices = numpy.arange(sl[0], sl[1], sl[2]).astype(numpy.int64)
            if isinstance(value, numpy.ndarray):
                values = value
            else:
                values = numpy.full(indices.shape, value)
            add_step = OnnxVar(self.inputs[0], indices, values,
                               op=OnnxScatterElements, axis=0)

        self.inputs = [add_step]
        return self

    def _setitem2i_(self, index, value):
        OnnxWhere = loadop('Where')
        add_step = OnnxVar(index, value, self.inputs[0], op=OnnxWhere)
        self.inputs = [add_step]
        return self

    def copy(self):
        """
        Returns a copy of self (use of Identity node).
        """
        OnnxIdentity = loadop('Identity')
        return OnnxVar(self, op=OnnxIdentity)

    def flatten(self, axis=0):
        """
        Flattens a matrix (see :epkg:`numpy:ndarray:flatten`).

        :param axis: only flatten from axis to the end.
        :return: @see cl OnnxVar.
        """
        OnnxFlatten, OnnxSqueeze = loadop('Flatten', 'Squeeze')
        fl = OnnxVar(self, op=OnnxFlatten, axis=axis)
        if axis == 0:
            return OnnxVar(fl, numpy.array([0], dtype=numpy.int64),
                           op=OnnxSqueeze)
        return fl


class MultiOnnxVar:
    """
    Class used to return multiple @see cl OnnxVar
    at the same time.
    """

    def __init__(self, *inputs, op=None, dtype=None, **kwargs):
        "constructor"
        logger.debug('MultiOnnxVar(%d in, dtype=%r, op=%r)',
                     len(inputs), dtype, op)
        self.onxvar = OnnxVar(*inputs, op=op, dtype=None, **kwargs)
        self.alg_ = None

    def _guess_dtype(self, dtype):
        "Guesses dtype when not specified."
        return self.onxvar._guess_dtype(dtype)

    @property
    def inputs(self):
        "Returns `self.onxvar.inputs`."
        return self.onxvar.inputs

    @property
    def onnx_op(self):
        "Returns `self.onxvar.onnx_op`."
        return self.onxvar.onnx_op

    @property
    def onnx_op_kwargs(self):
        "Returns `self.onxvar.onnx_op_kwargs`."
        return self.onxvar.onnx_op_kwargs

    def to_algebra(self, op_version=None):
        """
        Converts the variable into an operator.
        """
        if self.alg_ is None:
            logger.debug('MultiOnnxVar.to_algebra(op_version=%r)',
                         op_version)
            new_inputs = []
            for inp in self.inputs:
                if isinstance(inp, (
                        int, float, str, numpy.ndarray, numpy.int32,
                        numpy.int64, numpy.float32, numpy.float64,
                        numpy_bool, numpy_str, numpy.int8, numpy.uint8,
                        numpy.int16, numpy.uint16, numpy.uint32,
                        numpy.uint64)):
                    new_inputs.append(inp)
                elif hasattr(inp, 'fit'):
                    # scikit-learn models
                    new_inputs.append(inp)
                else:
                    new_inputs.append(
                        inp.to_algebra(op_version=op_version))

            if self.onnx_op is None:
                if len(new_inputs) == 1:
                    logger.debug('MultiOnnxVar.to_algebra:1:new_inputs[0]=%r',
                                 new_inputs[0])
                    self.alg_ = OnnxOperatorTuple(new_inputs[0])
                else:
                    logger.debug('MultiOnnxVar.to_algebra:2:new_inputs=%r',
                                 new_inputs)
                    self.alg_ = OnnxOperatorTuple(
                        new_inputs[0], *(new_inputs[1:]))
            else:
                logger.debug('MultiOnnxVar.to_algebra:%s:new_inputs=%r',
                             self.onnx_op.__class__.__name__, new_inputs)
                res = self.onnx_op(  # pylint: disable=E1102
                    *new_inputs, op_version=op_version, **self.onnx_op_kwargs)
                self.alg_ = OnnxOperatorTuple(res)
        return self.alg_

    def __getitem__(self, index):
        """
        Returns the ith elements.
        """
        return OnnxVar(self, index=index, op=OnnxOperatorItem)



# coding: utf-8
"""
@file
@brief Speeding up :epkg:`scikit-learn` with :epkg:`onnx`.

.. versionadded:: 0.7
"""
import collections
import inspect
import io
from contextlib import redirect_stdout, redirect_stderr
import numpy
from numpy.testing import assert_almost_equal
import scipy.special as scipy_special
import scipy.spatial.distance as scipy_distance
from onnx import helper, load
from sklearn.base import (
    BaseEstimator, clone,
    TransformerMixin, RegressorMixin, ClassifierMixin,
    ClusterMixin)
from sklearn.preprocessing import FunctionTransformer
from skl2onnx.algebra.onnx_operator_mixin import OnnxOperatorMixin
from ..tools.code_helper import print_code
from .. import __max_supported_opset__
from ..onnx_tools.onnx_export import export2numpy
from ..onnx_tools.onnx2py_helper import (
    onnx_model_opsets, _var_as_dict, to_skl2onnx_type)
from ..onnx_tools.exports.numpy_helper import (
    array_feature_extrator,
    argmax_use_numpy_select_last_index,
    argmin_use_numpy_select_last_index,
    make_slice)
from ..onnx_tools.exports.skl2onnx_helper import add_onnx_graph
from ..onnx_conv import to_onnx
from .onnx_transformer import OnnxTransformer


class _OnnxPipelineStepSpeedup(BaseEstimator, OnnxOperatorMixin):
    """
    Speeds up inference by replacing methods *transform* or
    *predict* by a runtime for :epkg:`ONNX`.

    :param estimator: estimator to train
    :param enforce_float32: boolean
        :epkg:`onnxruntime` only supports *float32*,
        :epkg:`scikit-learn` usually uses double floats, this parameter
        ensures that every array of double floats is converted into
        single floats
    :param runtime: string, defined the runtime to use
        as described in @see cl OnnxInference.
    :param target_opset: targetted ONNX opset
    :param conv_options: options for conversions, see @see fn to_onnx
    :param nopython: used by :epkg:`numba` jitter

    Attributes created by method *fit*:

    * `estimator_`: cloned and trained version of *estimator*
    * `onnxrt_`: objet of type @see cl OnnxInference,
        :epkg:`sklearn:preprocessing:FunctionTransformer`
    * `numpy_code_`: python code equivalent to the inference
        method if the runtime is `'numpy'` or `'numba'`
    * `onnx_io_names_`: dictionary, additional information
        if the runtime is `'numpy'` or `'numba'`

    .. versionadded:: 0.7
    """

    def __init__(self, estimator, runtime='python', enforce_float32=True,
                 target_opset=None, conv_options=None, nopython=True):
        BaseEstimator.__init__(self)
        self.estimator = estimator
        self.runtime = runtime
        self.enforce_float32 = enforce_float32
        self.target_opset = target_opset
        self.conv_options = conv_options
        self.nopython = nopython

    def _check_fitted_(self):
        if not hasattr(self, 'onnxrt_'):
            raise AttributeError(  # pragma: no cover
                "Object must be be fit.")

    def _to_onnx(self, fitted_estimator, inputs):
        """
        Converts an estimator inference into :epkg:`ONNX`.

        :param estimator: any estimator following :epkg:`scikit-learn` API
        :param inputs: example of inputs
        :return: ONNX
        """
        return to_onnx(
            self.estimator_, inputs, target_opset=self.target_opset,
            options=self.conv_options)

    def _build_onnx_runtime(self, onx):
        """
        Returns an instance of @see cl OnnxTransformer which
        executes the ONNX graph.

        :param onx: ONNX graph
        :param runtime: runtime type (see @see cl OnnxInference)
        :return: instance of @see cl OnnxInference
        """
        if self.runtime in ('numpy', 'numba'):
            return self._build_onnx_runtime_numpy(onx)
        tr = OnnxTransformer(
            onx, runtime=self.runtime,
            enforce_float32=self.enforce_float32)
        tr.fit()
        return tr

    def _build_onnx_runtime_numpy(self, onx):
        """
        Builds a runtime based on numpy.
        Exports the ONNX graph into python code
        based on numpy and then dynamically compiles
        it with method @see me _build_onnx_runtime_numpy_compile.
        """
        model_onnx = load(io.BytesIO(onx))
        self.onnx_io_names_ = {'inputs': [], 'outputs': []}
        for inp in model_onnx.graph.input:  # pylint: disable=E1101
            d = _var_as_dict(inp)
            self.onnx_io_names_['inputs'].append((d['name'], d['type']))
        for inp in model_onnx.graph.output:  # pylint: disable=E1101
            d = _var_as_dict(inp)
            self.onnx_io_names_['outputs'].append((d['name'], d['type']))
        self.onnx_io_names_['skl2onnx_inputs'] = [
            to_skl2onnx_type(d[0], d[1]['elem'], d[1]['shape'])
            for d in self.onnx_io_names_['inputs']]
        self.onnx_io_names_['skl2onnx_outputs'] = [
            to_skl2onnx_type(d[0], d[1]['elem'], d[1]['shape'])
            for d in self.onnx_io_names_['outputs']]
        self.numpy_code_ = export2numpy(model_onnx, rename=True)
        opsets = onnx_model_opsets(model_onnx)
        return self._build_onnx_runtime_numpy_compile(opsets)

    def _build_onnx_runtime_numpy_compile(self, opsets):
        """
        Second part of @see me _build_onnx_runtime_numpy.
        """
        try:
            compiled_code = compile(
                self.numpy_code_, '<string>', 'exec')
        except SyntaxError as e:  # pragma: no cover
            raise AssertionError(
                "Unable to compile a script due to %r. "
                "\n--CODE--\n%s"
                "" % (e, print_code(self.numpy_code_))) from e

        glo = globals().copy()
        loc = {
            'numpy': numpy, 'dict': dict, 'list': list,
            'print': print, 'sorted': sorted,
            'collections': collections, 'inspect': inspect,
            'helper': helper, 'scipy_special': scipy_special,
            'scipy_distance': scipy_distance,
            'array_feature_extrator': array_feature_extrator,
            'argmin_use_numpy_select_last_index':
                argmin_use_numpy_select_last_index,
            'argmax_use_numpy_select_last_index':
                argmax_use_numpy_select_last_index,
            'make_slice': make_slice}
        out = io.StringIO()
        err = io.StringIO()
        with redirect_stdout(out):
            with redirect_stderr(err):
                try:
                    exec(compiled_code, glo, loc)  # pylint: disable=W0122
                except Exception as e:  # pragma: no cover
                    raise AssertionError(
                        "Unable to execute a script due to %r. "
                        "\n--OUT--\n%s\n--ERR--\n%s\n--CODE--\n%s"
                        "" % (e, out.getvalue(), err.getvalue(),
                              print_code(self.numpy_code_))) from e
        names = [k for k in loc if k.startswith('numpy_')]
        if len(names) != 1:
            raise RuntimeError(  # pragma: no cover
                "Unable to guess which function is the one, names=%r."
                "" % list(sorted(names)))
        fct = loc[names[0]]
        if self.runtime == 'numba':
            from numba import jit
            jitter = jit(nopython=self.nopython)
            fct = jitter(fct)
        cl = FunctionTransformer(fct, accept_sparse=True)
        cl.op_version = opsets.get('', __max_supported_opset__)
        return cl

    def __getstate__(self):
        """
        :epkg:`pickle` does not support functions.
        This method removes any link to function
        when the runtime is `'numpy'`.
        """
        state = BaseEstimator.__getstate__(self)
        if 'numpy_code_' in state:
            del state['onnxrt_']
        return state

    def __setstate__(self, state):
        """
        :epkg:`pickle` does not support functions.
        This method restores the function created when
        the runtime is `'numpy'`.
        """
        BaseEstimator.__setstate__(self, state)
        if 'numpy_code_' in state:
            model_onnx = load(io.BytesIO(state['onnx_']))
            opsets = onnx_model_opsets(model_onnx)
            self.onnxrt_ = self._build_onnx_runtime_numpy_compile(opsets)

    def fit(self, X, y=None, sample_weight=None, **kwargs):
        """
        Fits the estimator, converts to ONNX.

        :param X: features
        :param args: other arguments
        :param kwargs: fitting options
        """
        if not hasattr(self, 'estimator_'):
            self.estimator_ = clone(self.estimator)
        if y is None:
            if sample_weight is None:
                self.estimator_.fit(X, **kwargs)
            else:
                self.estimator_.fit(X, sample_weight=sample_weight, **kwargs)
        else:
            if sample_weight is None:
                self.estimator_.fit(X, y, **kwargs)
            else:
                self.estimator_.fit(
                    X, y, sample_weight=sample_weight, **kwargs)

        if self.enforce_float32:
            X = X.astype(numpy.float32)
        self.onnx_ = self._to_onnx(self.estimator_, X).SerializeToString()
        self.onnxrt_ = self._build_onnx_runtime(self.onnx_)
        return self

    @property
    def op_version(self):
        """
        Returns the opset version.
        """
        self._check_fitted_()
        return self.onnxrt_.op_version

    def onnx_parser(self):
        """
        Returns a parser for this model.
        """
        self._check_fitted_()
        if isinstance(self.onnxrt_, FunctionTransformer):
            def parser():
                # Types should be included as well.
                return [r[0] for r in self.onnx_io_names_['skl2onnx_outputs']]
            return parser
        return self.onnxrt_.onnx_parser()

    def onnx_shape_calculator(self):
        """
        Returns a shape calculator for this transform.
        """
        self._check_fitted_()

        if isinstance(self.onnxrt_, FunctionTransformer):
            def fct_shape_calculator(operator):
                # Types should be included as well.
                outputs = self.onnx_io_names_['skl2onnx_outputs']
                if len(operator.outputs) != len(outputs):
                    raise RuntimeError(  # pragma: no cover
                        "Mismatch between parser and shape calculator, "
                        "%r != %r." % (outputs, operator.outputs))
                for a, b in zip(operator.outputs, outputs):
                    a.type = b[1]
            return fct_shape_calculator

        calc = self.onnxrt_.onnx_shape_calculator()

        def shape_calculator(operator):
            return calc(operator)

        return shape_calculator

    def onnx_converter(self):
        """
        Returns a converter for this transform.
        """
        self._check_fitted_()

        if isinstance(self.onnxrt_, FunctionTransformer):

            def fct_converter(scope, operator, container):
                op = operator.raw_operator
                onnx_model = load(io.BytesIO(op.onnx_))
                add_onnx_graph(scope, operator, container, onnx_model)

            return fct_converter

        conv = self.onnxrt_.onnx_converter()

        def converter(scope, operator, container):
            op = operator.raw_operator
            onnx_model = op.onnxrt_.onnxrt_.obj
            conv(scope, operator, container, onnx_model=onnx_model)

        return converter


class OnnxSpeedupTransformer(TransformerMixin,
                             _OnnxPipelineStepSpeedup):
    """
    Trains with :epkg:`scikit-learn`, transform with :epkg:`ONNX`.

    :param estimator: estimator to train
    :param enforce_float32: boolean
        :epkg:`onnxruntime` only supports *float32*,
        :epkg:`scikit-learn` usually uses double floats, this parameter
        ensures that every array of double floats is converted into
        single floats
    :param runtime: string, defined the runtime to use
        as described in @see cl OnnxInference.
    :param target_opset: targetted ONNX opset
    :param conv_options: conversion options, see @see fn to_onnx
    :param nopython: used by :epkg:`numba` jitter

    Attributes created by method *fit*:

    * `estimator_`: cloned and trained version of *estimator*
    * `onnxrt_`: objet of type @see cl OnnxInference,
        :epkg:`sklearn:preprocessing:FunctionTransformer`
    * `numpy_code_`: python code equivalent to the inference
        method if the runtime is `'numpy'` or `'numba'`
    * `onnx_io_names_`: dictionary, additional information
        if the runtime is `'numpy'` or `'numba'`

    .. versionadded:: 0.7
    """

    def __init__(self, estimator, runtime='python', enforce_float32=True,
                 target_opset=None, conv_options=None, nopython=True):
        _OnnxPipelineStepSpeedup.__init__(
            self, estimator, runtime=runtime, enforce_float32=enforce_float32,
            target_opset=target_opset, conv_options=conv_options,
            nopython=nopython)

    def fit(self, X, y=None, sample_weight=None):  # pylint: disable=W0221
        """
        Trains based estimator.
        """
        if sample_weight is None:
            _OnnxPipelineStepSpeedup.fit(self, X, y)
        else:
            _OnnxPipelineStepSpeedup.fit(
                self, X, y, sample_weight=sample_weight)
        return self

    def transform(self, X):
        """
        Transforms with *ONNX*.

        :param X: features
        :return: transformed features
        """
        return self.onnxrt_.transform(X)

    def raw_transform(self, X):
        """
        Transforms with *scikit-learn*.

        :param X: features
        :return: transformed features
        """
        return self.estimator_.transform(X)

    def assert_almost_equal(self, X, **kwargs):
        """
        Checks that ONNX and scikit-learn produces the same
        outputs.
        """
        expected = self.raw_transform(X)
        got = self.transform(X)
        assert_almost_equal(expected, got, **kwargs)


class OnnxSpeedupRegressor(RegressorMixin,
                           _OnnxPipelineStepSpeedup):
    """
    Trains with :epkg:`scikit-learn`, transform with :epkg:`ONNX`.

    :param estimator: estimator to train
    :param enforce_float32: boolean
        :epkg:`onnxruntime` only supports *float32*,
        :epkg:`scikit-learn` usually uses double floats, this parameter
        ensures that every array of double floats is converted into
        single floats
    :param runtime: string, defined the runtime to use
        as described in @see cl OnnxInference.
    :param target_opset: targetted ONNX opset
    :param conv_options: conversion options, see @see fn to_onnx
    :param nopython: used by :epkg:`numba` jitter

    Attributes created by method *fit*:

    * `estimator_`: cloned and trained version of *estimator*
    * `onnxrt_`: objet of type @see cl OnnxInference,
        :epkg:`sklearn:preprocessing:FunctionTransformer`
    * `numpy_code_`: python code equivalent to the inference
        method if the runtime is `'numpy'` or `'numba'`
    * `onnx_io_names_`: dictionary, additional information
        if the runtime is `'numpy'` or `'numba'`

    .. versionadded:: 0.7
    """

    def __init__(self, estimator, runtime='python', enforce_float32=True,
                 target_opset=None, conv_options=None, nopython=True):
        _OnnxPipelineStepSpeedup.__init__(
            self, estimator, runtime=runtime, enforce_float32=enforce_float32,
            target_opset=target_opset, conv_options=conv_options,
            nopython=nopython)

    def fit(self, X, y, sample_weight=None):  # pylint: disable=W0221
        """
        Trains based estimator.
        """
        if sample_weight is None:
            _OnnxPipelineStepSpeedup.fit(self, X, y)
        else:
            _OnnxPipelineStepSpeedup.fit(
                self, X, y, sample_weight=sample_weight)
        return self

    def predict(self, X):
        """
        Transforms with *ONNX*.

        :param X: features
        :return: transformed features
        """
        return self.onnxrt_.transform(X)

    def raw_predict(self, X):
        """
        Transforms with *scikit-learn*.

        :param X: features
        :return: transformed features
        """
        return self.estimator_.predict(X)

    def assert_almost_equal(self, X, **kwargs):
        """
        Checks that ONNX and scikit-learn produces the same
        outputs.
        """
        expected = numpy.squeeze(self.raw_predict(X))
        got = numpy.squeeze(self.predict(X))
        assert_almost_equal(expected, got, **kwargs)


class OnnxSpeedupClassifier(ClassifierMixin,
                            _OnnxPipelineStepSpeedup):
    """
    Trains with :epkg:`scikit-learn`, transform with :epkg:`ONNX`.

    :param estimator: estimator to train
    :param enforce_float32: boolean
        :epkg:`onnxruntime` only supports *float32*,
        :epkg:`scikit-learn` usually uses double floats, this parameter
        ensures that every array of double floats is converted into
        single floats
    :param runtime: string, defined the runtime to use
        as described in @see cl OnnxInference.
    :param target_opset: targetted ONNX opset
    :param conv_options: conversion options, see @see fn to_onnx
    :param nopython: used by :epkg:`numba` jitter

    Attributes created by method *fit*:

    * `estimator_`: cloned and trained version of *estimator*
    * `onnxrt_`: objet of type @see cl OnnxInference,
        :epkg:`sklearn:preprocessing:FunctionTransformer`
    * `numpy_code_`: python code equivalent to the inference
        method if the runtime is `'numpy'` or `'numba'`
    * `onnx_io_names_`: dictionary, additional information
        if the runtime is `'numpy'` or `'numba'`

    .. versionadded:: 0.7
    """

    def __init__(self, estimator, runtime='python', enforce_float32=True,
                 target_opset=None, conv_options=None, nopython=True):
        if conv_options is None:
            conv_options = {'zipmap': False}
        _OnnxPipelineStepSpeedup.__init__(
            self, estimator, runtime=runtime, enforce_float32=enforce_float32,
            target_opset=target_opset, conv_options=conv_options,
            nopython=nopython)

    def fit(self, X, y, sample_weight=None):  # pylint: disable=W0221
        """
        Trains based estimator.
        """
        if sample_weight is None:
            _OnnxPipelineStepSpeedup.fit(self, X, y)
        else:
            _OnnxPipelineStepSpeedup.fit(
                self, X, y, sample_weight=sample_weight)
        return self

    def predict(self, X):
        """
        Transforms with *ONNX*.

        :param X: features
        :return: transformed features
        """
        pred = self.onnxrt_.transform(X)
        if isinstance(pred, tuple):
            return pred[0]
        return pred.iloc[:, 0].values

    def predict_proba(self, X):
        """
        Transforms with *ONNX*.

        :param X: features
        :return: transformed features
        """
        pred = self.onnxrt_.transform(X)
        if isinstance(pred, tuple):
            return pred[1]
        return pred.iloc[:, 1:].values

    def raw_predict(self, X):
        """
        Transforms with *scikit-learn*.

        :param X: features
        :return: transformed features
        """
        return self.estimator_.predict(X)

    def raw_predict_proba(self, X):
        """
        Transforms with *scikit-learn*.

        :param X: features
        :return: transformed features
        """
        return self.estimator_.predict_proba(X)

    def assert_almost_equal(self, X, **kwargs):
        """
        Checks that ONNX and scikit-learn produces the same
        outputs.
        """
        expected = numpy.squeeze(self.raw_predict_proba(X))
        got = numpy.squeeze(self.predict_proba(X))
        assert_almost_equal(expected, got, **kwargs)
        expected = numpy.squeeze(self.raw_predict(X))
        got = numpy.squeeze(self.predict(X))
        assert_almost_equal(expected, got, **kwargs)


class OnnxSpeedupCluster(ClusterMixin,
                         _OnnxPipelineStepSpeedup):
    """
    Trains with :epkg:`scikit-learn`, transform with :epkg:`ONNX`.

    :param estimator: estimator to train
    :param enforce_float32: boolean
        :epkg:`onnxruntime` only supports *float32*,
        :epkg:`scikit-learn` usually uses double floats, this parameter
        ensures that every array of double floats is converted into
        single floats
    :param runtime: string, defined the runtime to use
        as described in @see cl OnnxInference.
    :param target_opset: targetted ONNX opset
    :param conv_options: conversion options, see @see fn to_onnx
    :param nopython: used by :epkg:`numba` jitter

    Attributes created by method *fit*:

    * `estimator_`: cloned and trained version of *estimator*
    * `onnxrt_`: objet of type @see cl OnnxInference,
        :epkg:`sklearn:preprocessing:FunctionTransformer`
    * `numpy_code_`: python code equivalent to the inference
        method if the runtime is `'numpy'` or `'numba'`
    * `onnx_io_names_`: dictionary, additional information
        if the runtime is `'numpy'` or `'numba'`

    .. versionadded:: 0.7
    """

    def __init__(self, estimator, runtime='python', enforce_float32=True,
                 target_opset=None, conv_options=None, nopython=True):
        _OnnxPipelineStepSpeedup.__init__(
            self, estimator, runtime=runtime, enforce_float32=enforce_float32,
            target_opset=target_opset, conv_options=conv_options,
            nopython=nopython)

    def fit(self, X, y, sample_weight=None):  # pylint: disable=W0221
        """
        Trains based estimator.
        """
        if sample_weight is None:
            _OnnxPipelineStepSpeedup.fit(self, X, y)
        else:
            _OnnxPipelineStepSpeedup.fit(
                self, X, y, sample_weight=sample_weight)
        return self

    def predict(self, X):
        """
        Transforms with *ONNX*.

        :param X: features
        :return: transformed features
        """
        pred = self.onnxrt_.transform(X)
        if isinstance(pred, tuple):
            return pred[0]
        return pred.iloc[:, 0].values

    def transform(self, X):
        """
        Transforms with *ONNX*.

        :param X: features
        :return: transformed features
        """
        pred = self.onnxrt_.transform(X)
        if isinstance(pred, tuple):
            return pred[1]
        return pred.iloc[:, 1:].values

    def raw_predict(self, X):
        """
        Transforms with *scikit-learn*.

        :param X: features
        :return: transformed features
        """
        return self.estimator_.predict(X)

    def raw_transform(self, X):
        """
        Transforms with *scikit-learn*.

        :param X: features
        :return: transformed features
        """
        return self.estimator_.transform(X)

    def assert_almost_equal(self, X, **kwargs):
        """
        Checks that ONNX and scikit-learn produces the same
        outputs.
        """
        expected = numpy.squeeze(self.raw_transform(X))
        got = numpy.squeeze(self.transform(X))
        assert_almost_equal(expected, got, **kwargs)
        expected = numpy.squeeze(self.raw_predict(X))
        got = numpy.squeeze(self.predict(X))
        assert_almost_equal(expected, got, **kwargs)



"""
@file
@brief Custom operators for gradient numbers.
"""
from skl2onnx.algebra.onnx_operator import OnnxOperator


class OnnxYieldOp_1(OnnxOperator):
    """
    Defines a custom operator for YieldOp.
    """

    since_version = 1
    expected_inputs = [('X', 'T')]
    expected_outputs = [('Y', 'T')]
    input_range = [1, 1]
    output_range = [1, 1]
    is_deprecated = False
    domain = 'com.microsoft'
    operator_name = 'YieldOp'
    past_version = {}

    def __init__(self, X, non_differentiable_outputs=None,
                 full_shape_outputs=None, op_version=None, **kwargs):
        """
        :param X: array or OnnxOperatorMixin
        :param non_differentiable_outputs: the indices of the module
            outputs that doesn't have a gradient.
        :param full_shape_outputs: the indices of the module outputs
            that must have full shape.
        :param op_version: opset version
        :param kwargs: additional parameter
        """
        OnnxOperator.__init__(
            self, X, op_version=op_version, **kwargs)
        self.non_differentiable_outputs = non_differentiable_outputs
        self.full_shape_outputs = full_shape_outputs


OnnxYieldOp = OnnxYieldOp_1


class OnnxBroadcastGradientArgs_1(OnnxOperator):
    """
    Defines a custom operator for BroadcastGradientArgs.
    Returns the reduction axes for computing gradients of s0 op s1 with
    broadcast. The ouput axes are deterministic from last to first.
    Output is an empty vector when no reduction is necessary for the
    corresponding input.
    """

    since_version = 1
    expected_inputs = [('a_shape', 'T'), ('b_shape', 'T')]
    expected_outputs = [('a_axes', 'T'), ('b_axes', 'T')]
    input_range = [2, 2]
    output_range = [2, 2]
    is_deprecated = False
    domain = 'com.microsoft'
    operator_name = 'BroadcastGradientArgs'
    past_version = {}

    def __init__(self, a_shape, b_shape, op_version=None, **kwargs):
        """
        :param a_shape: The 1st input shape as Tensor.
        :param b_shape: The 2nds input shape as Tensor.
        :param op_version: opset version
        :param kwargs: additional parameter
        """
        OnnxOperator.__init__(
            self, a_shape, b_shape, op_version=op_version, **kwargs)


OnnxBroadcastGradientArgs = OnnxBroadcastGradientArgs_1


class OnnxFusedMatMul_1(OnnxOperator):
    """
    MatMul and Gemm without a C.
    """

    since_version = 1
    expected_inputs = [('X', 'T'), ('X', 'T')]
    expected_outputs = [('Z', 'T')]
    input_range = [2, 2]
    output_range = [1, 1]
    is_deprecated = False
    domain = 'com.microsoft'
    operator_name = 'FusedMatMul'
    past_version = {}

    def __init__(self, X, Y, transA=0, transB=0,
                 op_version=None, **kwargs):
        """
        :param X: first matrix
        :param Y: second matrix
        :param transA: transpose first matrix
        :param transB: transpose second matrix
        :param op_version: opset version
        :param kwargs: additional parameter
        """
        OnnxOperator.__init__(
            self, X, Y, transA=transA, transB=transB,
            op_version=op_version, **kwargs)


OnnxFusedMatMul = OnnxFusedMatMul_1


class OnnxSoftmaxGrad_13(OnnxOperator):
    """
    Gradient of Softmax.
    SoftmaxGrad computes :math:`Y * ( dY - ReduceSum(Y * dY))`.
    ONNX does not have a dot product,
    which can be simulated as a pointwise-multiplication ("Mul"),
    followed by a "ReduceSum". Unfortunately, the treatment of "axis"
    is different in "SoftmaxGrad" and "ReduceSum".
    If axis=k for SoftmaxGrad, we need to specify [k, ..., n-1] as the axes of
    reduction for "ReduceSum", after accounting for negative-axis specification.
    An alternative solution would be to Flatten inputs to 2D and then reshape
    output back to original shape. Hopefully, many of these ops can be optimized
    away in the common-case of statically-known shapes.
    """

    since_version = 1
    expected_inputs = [('grad', 'T'), ('prob', 'T')]
    expected_outputs = [('Y', 'T')]
    input_range = [2, 2]
    output_range = [1, 1]
    is_deprecated = False
    domain = 'com.microsoft'
    operator_name = 'SoftmaxGrad_13'
    past_version = {}

    def __init__(self, grad, prob, op_version=None, **kwargs):
        """
        :param grad: gradient
        :param prob: probablities
        :param op_version: opset version
        :param kwargs: additional parameter
        """
        OnnxOperator.__init__(
            self, grad, prob, op_version=op_version, **kwargs)


OnnxSoftmaxGrad = OnnxSoftmaxGrad_13



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class Size(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc, **options)

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (numpy.array(data.size, dtype=numpy.int64), )



"""
@file
@brief Command line about model optimisation.
"""
import os
import onnx


def onnx_stats(name, optim=False, kind=None):
    """
    Computes statistics on an ONNX model.

    :param name: filename
    :param optim: computes statistics before an after optimisation was done
    :param kind: kind of statistics, if left unknown,
        prints out the metadata, possible values:
        * `io`: prints input and output name, type, shapes
        * `node`: prints the distribution of node types
        * `text`: printts a text summary

    .. cmdref::
        :title: Computes statistics on an ONNX graph
        :cmd: -m mlprodict onnx_stats --help
        :lid: l-cmd-onnx_stats

        The command computes statistics on an ONNX model.
    """
    if not os.path.exists(name):
        raise FileNotFoundError(  # pragma: no cover
            f"Unable to find file '{name}'.")
    with open(name, 'rb') as f:
        model = onnx.load(f)
    if kind in (None, ""):
        from ..onnx_tools.optim import onnx_statistics
        return onnx_statistics(model, optim=optim)
    if kind == 'text':
        from ..plotting.plotting import onnx_simple_text_plot
        return onnx_simple_text_plot(model)
    if kind == 'io':
        from ..plotting.plotting import onnx_text_plot_io
        return onnx_text_plot_io(model)
    if kind == 'node':
        from ..onnx_tools.optim import onnx_statistics
        return onnx_statistics(model, optim=optim, node_type=True)
    raise ValueError(  # pragma: no cover
        f"Unexpected kind={kind!r}.")


def onnx_optim(name, outfile=None, recursive=True, options=None, verbose=0, fLOG=None):
    """
    Optimizes an ONNX model.

    :param name: filename
    :param outfile: output filename
    :param recursive: processes the main graph and the subgraphs
    :param options: options, kind of optimize to do
    :param verbose: display statistics before and after the optimisation
    :param fLOG: logging function

    .. cmdref::
        :title: Optimizes an ONNX graph
        :cmd: -m mlprodict onnx_optim --help
        :lid: l-cmd-onnx_optim

        The command optimizes an ONNX model.
    """
    from ..onnx_tools.optim import onnx_statistics, onnx_optimisations
    if not os.path.exists(name):
        raise FileNotFoundError(  # pragma: no cover
            f"Unable to find file '{name}'.")
    if outfile == "":
        outfile = None  # pragma: no cover
    if options == "":
        options = None  # pragma: no cover
    if verbose >= 1 and fLOG is not None:
        fLOG(f"[onnx_optim] read file '{name}'.")
    with open(name, 'rb') as f:
        model = onnx.load(f)
    if verbose >= 1 and fLOG is not None:
        stats = onnx_statistics(model, optim=False)
        for k, v in sorted(stats.items()):
            fLOG(f'  before.{k}={v}')
    new_model = onnx_optimisations(model, recursive=recursive)
    if verbose >= 1 and fLOG is not None:
        stats = onnx_statistics(model, optim=False)
        for k, v in sorted(stats.items()):
            fLOG(f'   after.{k}={v}')
    if outfile is not None:
        fLOG(f"[onnx_optim] write '{outfile}'.")
        with open(outfile, 'wb') as f:
            onnx.save(new_model, f)
    return new_model



# pylint: disable=C0302
"""
@file
@brief Classes representing the sequence of matrix operations to
implement einsum computation.
"""
import numpy
from onnx import helper, numpy_helper
from ...onnx_tools.onnx2py_helper import guess_proto_dtype
from ...npy.xop_variable import guess_numpy_type
from ... import __max_supported_opset__, get_ir_version
from .blas_lapack import gemm_dot
from .einsum_impl_ext import (
    numpy_extended_dot, numpy_diagonal,
    _numpy_extended_dot_equation,
    numpy_extended_dot_python,
    numpy_extended_dot_matrix)


def single_axes(axes):
    """
    *axes* contains positive values, then it is the position
    of this axis in the original matrix, otherwise it is -1
    meaning this axis is an added single dimension to align
    all the dimensions based on the einsum equation.

    :param axes: axes described above
    :return: list of integer in set `{1, 2}`, 1 for
        a single axis, 2 otherwise
    """
    if axes is None:
        return axes
    return [(1 if a == -1 else 2) for a in axes]


class EinsumSubOp:
    """
    Defines a sub operation used in Einsum decomposition.

    :param name: name (reshape, transpose, reduce_sum, matmul, id,
        squeeze, diagonal, mul, batch_dot)
    :param inputs: inputs
    :param kwargs: arguments

    Operator suffixed by `_mm` (*transpose_mm*, *reduce_sum_mm*)
    are equivalent to the same operator without the suffix
    but takes two inputs and only changes the first one.

    Attributes `_info` summarizes the known information
    about dimensions. Many of them are empty because inserted.
    Value `1` means it was the case, `2` means it is a plain dimension.
    """
    _allowed = {'expand_dims', 'transpose', 'reduce_sum', 'matmul', 'id',
                'squeeze', 'diagonal', 'mul', 'batch_dot',
                'transpose_mm', 'reduce_sum_mm'}

    def __init__(self, full_dim, name, *inputs, **kwargs):
        self.full_dim = full_dim
        self.name = name
        self.inputs = inputs
        self.kwargs = kwargs
        self._info = {}
        if name not in EinsumSubOp._allowed:
            raise ValueError(
                f"Unexpected name {name!r}. It should be in {EinsumSubOp._allowed!r}.")
        if len(inputs) not in (1, 2):
            raise RuntimeError(
                f"Inputs must contains 1 or 2 inputs not {len(inputs)}.")
        if name == 'matmul' and len(inputs) != 2:
            raise RuntimeError(
                "Inputs must contains 2 inputs not %d for operator 'matmul'."
                "" % len(inputs))
        for i, inp in enumerate(inputs):
            if not isinstance(inp, (int, EinsumSubOp)):
                raise TypeError(
                    "Input %d has type %r, int or EinsumSubOp is expected."
                    "" % (i, type(inp)))
        self._check_()

    def _check_(self):
        if self.name == 'transpose':
            self._check_arg_('perm', tuple)
            perm = self.kwargs['perm']
            if len(perm) != len(set(perm)):
                raise RuntimeError(  # pragma: no cover
                    f"perm has duplicated values {perm!r} (name={self.name!r}).")
            if list(perm) == list(range(len(perm))):
                raise ValueError(  # pragma: no cover
                    f"Transpose = identity perm={perm}. It must be removed.")
        elif self.name == 'matmul':
            self._check_arg_('axes', tuple)
            self._check_arg_('left', tuple)
            self._check_arg_('right', tuple)
            axes = self.kwargs['axes']
            left = self.kwargs['left']
            right = self.kwargs['right']
            for a in axes:
                if a in left and a in right:
                    raise RuntimeError(  # pragma: no cover
                        "One axis belongs to every set (axes, left, right). "
                        "axes=%r, left=%r, right=%r." % (axes, left, right))

    def __repr__(self):
        inps = ", ".join(map(str, self.inputs))
        kw = ", ".join(f"{k}={w!r}" for k, w in self.kwargs.items())
        m = f"{self.__class__.__name__}({self.name!r}, {inps}, {kw})"
        return m

    def dot_label(self):
        """
        Displays some informations useful to understand the operator.
        """
        if self.name == "matmul":
            ndim = self.kwargs['ndim']
            axes = self.kwargs['axes']
            left = self.kwargs['left']
            right = self.kwargs['right']
            eq = _numpy_extended_dot_equation(ndim, ndim, axes, left, right)
            eq = eq.replace(">", "\\\\>")
            return "~" + eq
        return None

    def _check_arg_(self, name, typ, empty=False):
        if name not in self.kwargs:
            raise RuntimeError(  # pragma: no cover
                f"Parameter {name!r} not found for operator {self.name!r}.")
        if empty and self.kwargs[name] is None:
            return
        if not isinstance(self.kwargs[name], typ):
            raise TypeError(  # pragma: no cover
                "Unexpected type %r for parameter %r and parameter %r."
                "" % (type(self.kwargs[name]), name, self.name))

    def _check_row_(self, row, inp=False, verbose=False):
        """
        Checks input or output is valid.
        """
        if verbose:
            if inp:
                print('<<' if inp else '>>', self.name, row, self.kwargs)
            else:
                print('<<' if inp else '>>', self.name, row)

    def _compute_output_row_id(self, row, row2=None, ab=False, verbose=False):
        if ab:
            raise RuntimeError("ab option not allowed.")  # pragma: no cover
        self._check_row_(row, True, verbose=verbose)
        row[:] = row2[:]
        self._check_row_(row, verbose=verbose)

    def _compute_output_row_transpose(self, row, row2=None, ab=False, verbose=False):
        if ab:
            self._compute_output_row_transpose(row2, verbose=verbose)
            return
        self._check_row_(row, True, verbose=verbose)
        self._check_arg_('perm', tuple)
        if len(self.kwargs['perm']) != len(row):
            raise RuntimeError(  # pragma: no cover
                f"Unexpected permutation {self.kwargs['perm']!r} (row={row!r}).")
        perm = self.kwargs['perm']
        cpy = row.copy()
        for i, p in enumerate(perm):
            row[i] = cpy[p]
        self._check_row_(row, verbose=verbose)

    def _compute_output_row_transpose_mm(self, row, row2=None, ab=False, verbose=False):
        if not ab:
            raise RuntimeError("ab must be True.")  # pragma: no cover
        self._check_row_(row, True, verbose=verbose)
        if row2 is None:
            raise RuntimeError(  # pragma: no cover
                "transpose_mm expects a second input.")
        self._compute_output_row_transpose(row, row2=None, verbose=verbose)

    def _compute_output_row_expand_dims(self, row, row2=None, ab=False, verbose=False):
        if ab:
            raise RuntimeError("ab option not allowed.")  # pragma: no cover
        self._check_row_(row, True, verbose=verbose)
        self._check_arg_('axes', tuple)
        axes = self.kwargs['axes']
        for axis in axes:
            if not isinstance(axis, tuple):
                raise TypeError(  # pragma: no cover
                    "Parameter axes of expand_dims should be a tuple of "
                    "tuple, axes=%r." % axes)
            if row[axis[1]] != -1:
                raise RuntimeError(  # pragma: no cover
                    "Dimension should be -1 in row %r axis=%r." % (
                        row, self.kwargs['axis']))
        self._check_row_(row, verbose=verbose)

    def _compute_output_row_reduce_sum(self, row, row2=None, ab=False, verbose=False):
        if ab:
            raise RuntimeError("ab option not allowed.")  # pragma: no cover
        self._check_row_(row, True, verbose=verbose)
        self._check_arg_('axes', tuple)
        for a in self.kwargs['axes']:
            row[a] = -1
        self._check_row_(row, verbose=verbose)

    def _compute_output_row_reduce_sum_mm(self, row, row2=None, ab=False, verbose=False):
        if not ab:
            raise RuntimeError("ab must be true.")  # pragma: no cover
        self._check_row_(row2, True, verbose=verbose)
        if row2 is None:
            raise RuntimeError(  # pragma: no cover
                "reduce_sum_mm expects a second input.")
        self._compute_output_row_reduce_sum(row, row2=None, verbose=verbose)

    def _compute_output_row_squeeze(self, row, row2=None, ab=False, verbose=False):
        if ab:
            raise RuntimeError("ab option not allowed.")  # pragma: no cover
        self._check_row_(row, True, verbose=verbose)
        self._check_arg_('axes', tuple)
        for a in self.kwargs['axes']:
            row[a] = -1
        self._check_row_(row, verbose=verbose)

    def _compute_output_row_diagonal(self, row, row2=None, ab=False, verbose=False):
        if ab:
            raise RuntimeError("ab option not allowed.")  # pragma: no cover
        self._check_row_(row, True, verbose=verbose)
        self._check_arg_('diag', list)
        to_remove = []
        for choice, choices in self.kwargs['diag']:
            for ch in choices:
                if ch != choice:
                    to_remove.append(ch)
            for i in range(len(row)):  # pylint: disable=C0200
                if row[i] in choices:
                    if row[i] != choice:
                        row[i] = choice
        to_remove.sort()
        for r in to_remove:
            for i in range(len(row)):  # pylint: disable=C0200
                if row[i] == r:
                    raise RuntimeError(  # pragma: no cover
                        "Unexpected result r=%r row=%r to_remove=%r "
                        "diag=%r." % (
                            r, row, to_remove, self.kwargs['diag']))
                if row[i] > r:
                    row[i] -= 1
        self._check_row_(row, verbose=verbose)

    def _compute_output_row_matmul(self, row, row2=None, ab=False, verbose=False):
        if not ab:
            raise RuntimeError("ab must be True.")  # pragma: no cover
        self._check_row_(row, True, verbose=verbose)
        self._check_row_(row2, True, verbose=verbose)
        self._check_arg_('axes', tuple)
        self._check_arg_('left', tuple)
        self._check_arg_('right', tuple)
        self._check_arg_('ndim', int)
        if row2 is None:
            raise RuntimeError(  # pragma: no cover
                "matmul expects two inputs.")
        if verbose:
            ndim = self.kwargs['ndim']
            axes = self.kwargs['axes']
            left = self.kwargs['left']
            right = self.kwargs['right']
            print("    MATMUL %r @ %r axes=%r left=%r right=%r - eq=%s" % (
                row, row2, axes, left, right,
                _numpy_extended_dot_equation(ndim, ndim, axes, left, right)))
        row2[:] = numpy.maximum(row, row2)
        for a in self.kwargs['axes']:
            if a not in self.kwargs['right']:
                row2[a] = -1
        self._check_row_(row2, verbose=verbose)

    def _compute_output_row_batch_dot(self, row, row2=None, ab=False, verbose=False):
        if not ab:
            raise RuntimeError("ab must be True.")  # pragma: no cover
        self._check_row_(row, True, verbose=verbose)
        self._check_row_(row2, True, verbose=verbose)
        self._check_arg_('batch_axes', tuple)
        self._check_arg_('keep_axes', tuple, empty=True)
        self._check_arg_('sum_axes', tuple)
        self._check_arg_('left', tuple)
        self._check_arg_('right', tuple)
        self._check_arg_('ndim', int)
        if row2 is None:
            raise RuntimeError(
                "batch_dot expects two inputs.")  # pragma: no cover
        if verbose:
            batch_axes = self.kwargs['batch_axes']
            keep_axes = self.kwargs['keep_axes']
            sum_axes = self.kwargs['sum_axes']
            left = self.kwargs['left']
            right = self.kwargs['right']
            ndim = self.kwargs['ndim']
            print("    BATCH_DOT batch_axes=%r keep_axes=%r sum_axes=%r "
                  "left=%r right=%r eq=%r" % (
                      batch_axes, keep_axes, sum_axes, left, right,
                      _numpy_extended_dot_equation(ndim, ndim, sum_axes, left, right)))
        row2[:] = numpy.maximum(row, row2)
        for a in self.kwargs['sum_axes']:
            if a not in self.kwargs['right']:
                row2[a] = -1
        self._check_row_(row2, verbose=verbose)

    def _compute_output_row_mul(self, row, row2=None, ab=False, verbose=False):
        if not ab:
            raise RuntimeError("ab must be True.")  # pragma: no cover
        self._check_row_(row, True, verbose=verbose)
        self._check_row_(row2, True, verbose=verbose)
        if row2 is None:
            raise RuntimeError("mul expects two inputs.")  # pragma: no cover
        if verbose:
            print(  # pragma: no cover
                f"    MUL {row!r} @ {row2!r}")
        row2[:] = numpy.maximum(row, row2)
        self._check_row_(row2, verbose=verbose)

    def compute_output_row(self, row, row2=None, ab=False, verbose=False):
        """
        Updates *row* based on the operator.
        """
        method_name = f"_compute_output_row_{self.name}"
        meth = getattr(self, method_name, None)
        if meth is None:
            raise NotImplementedError(  # pragma: no cover
                f"compute_output_row not implemented for {self.name!r}.")
        if verbose and ab:
            print("  -- called as a binary operator")
        self.add_info(i_row=single_axes(row), i_row2=single_axes(row2))
        meth(row, row2=row2, ab=ab, verbose=verbose)
        self.add_info(o_row=single_axes(row), o_row2=single_axes(row2))

    def add_info(self, **kwargs):
        """
        Adds information to the node.

        :param kwargs: dictionary
        """
        for k, v in kwargs.items():
            if k in self._info:
                raise KeyError(  # pragma: no cover
                    f"Key {k!r} already added (operator {self.name!r}).")
            self._info[k] = v

    def _check_inputs_(self, n_expected, check_dim=False):
        if len(self.inputs) != n_expected:
            raise RuntimeError(  # pragma: no cover
                "Number of inputs must be %d not %d for operator %r."
                "" % (n_expected, len(self.inputs), self.name))

    def _check_shape_(self, m):
        if len(m.shape) != self.full_dim:
            raise RuntimeError(  # pragma: no cover
                "Number of dimensions %r is different from expected value "
                "%d." % (m.shape, self.full_dim))

    def _get_data(self, data, key):
        if isinstance(key, int):
            if key not in data:
                raise RuntimeError(  # pragma: no cover
                    "Unable to find key %d in %r." % (
                        key, list(sorted(data))))
            return data[key]
        if isinstance(key, EinsumSubOp):
            if id(key) not in data:
                raise RuntimeError(  # pragma: no cover
                    "Unable to find key %d in %r." % (
                        id(key), list(sorted(data))))
            return data[id(key)]
        raise TypeError(  # pragma: no cover
            f"Unexpected input type {type(key)!r}.")

    def _apply_id(self, data, verbose=False, **kwargs):
        self._check_inputs_(1)
        inp = self.inputs[0]
        output = self._get_data(data, inp)
        return output

    def _apply_diagonal(self, data, verbose=False, **kwargs):
        self._check_inputs_(1)
        inp = self.inputs[0]
        m = self._get_data(data, inp)
        if verbose:
            print(  # pragma: no cover
                f"- {self.name}, shape={m.shape!r} diag={self.kwargs['diag']!r}")
        diag = self.kwargs['diag']
        if len(diag) != 1:
            raise NotImplementedError(  # pragma: no cover
                f"Not implemented with more than one duplicated indice {diag!r}.")
        diag0 = diag[0]
        output = numpy_diagonal(m, axis=diag0[0], axes=diag0[1])
        return output

    def _apply_expand_dims(self, data, verbose=False, **kwargs):
        self._check_inputs_(1)
        inp = self.inputs[0]
        m = self._get_data(data, inp)
        if verbose:
            print(
                f"- {self.name}, shape={m.shape!r} axes={self.kwargs['axes']!r}")
        output = m
        for axis in reversed(self.kwargs['axes']):
            output = numpy.expand_dims(output, axis[0])
        return output

    def _apply_transpose(self, data, verbose=False, **kwargs):
        self._check_inputs_(1, True)
        inp = self.inputs[0]
        m = self._get_data(data, inp)
        self._check_shape_(m)
        if verbose:
            print(
                f"- {self.name}, shape={m.shape!r} perm={self.kwargs['perm']!r}")
        output = numpy.transpose(m, self.kwargs['perm'])
        self._check_shape_(output)
        return output

    def _apply_transpose_mm(self, data, verbose=False, **kwargs):
        self._check_inputs_(2, True)
        inp = self.inputs[0]
        m = self._get_data(data, inp)
        self._check_shape_(m)
        if verbose:
            print(  # pragma: no cover
                f"- {self.name}, shape={m.shape!r} perm={self.kwargs['perm']!r}")
        output = numpy.transpose(m, self.kwargs['perm'])
        self._check_shape_(output)
        return output

    def _apply_matmul(self, data, verbose=False, **kwargs):
        self._check_inputs_(2)
        inp1 = self.inputs[0]
        inp2 = self.inputs[1]
        m1 = self._get_data(data, inp1)
        m2 = self._get_data(data, inp2)
        self._check_shape_(m1)
        self._check_shape_(m2)
        axes = self.kwargs['axes']
        left = self.kwargs['left']
        right = self.kwargs['right']

        if verbose:
            print("- %s, shapes=%r @ %r axes=%r left=%r right=%r" % (
                self.name, m1.shape, m2.shape, axes, left, right))

        impl = kwargs.get('matmul_impl', None)
        if impl == 'pyf':
            output = numpy_extended_dot_matrix(m1, m2, axes, left, right,
                                               verbose=verbose)
        elif impl == 'py':
            output = numpy_extended_dot_python(m1, m2, axes, left, right,
                                               verbose=verbose)
        elif impl is None:
            output = numpy_extended_dot(m1, m2, axes, left, right,
                                        verbose=verbose)
        else:
            raise ValueError(
                f"Unknown implementation of numpy_extended_dot ({impl}).")
        self._check_shape_(output)
        return output

    def _apply_mul(self, data, verbose=False, **kwargs):
        self._check_inputs_(2)
        inp1 = self.inputs[0]
        inp2 = self.inputs[1]
        m1 = self._get_data(data, inp1)
        m2 = self._get_data(data, inp2)
        self._check_shape_(m1)
        self._check_shape_(m2)

        if verbose:
            print(  # pragma: no cover
                f"- {self.name}, shapes={m1.shape!r} @ {m2.shape!r}")

        output = m1 * m2
        self._check_shape_(output)
        return output

    def _apply_batch_dot(self, data, verbose=False, **kwargs):
        self._check_inputs_(2)
        inp1 = self.inputs[0]
        inp2 = self.inputs[1]
        m1 = self._get_data(data, inp1)
        m2 = self._get_data(data, inp2)
        self._check_shape_(m1)
        self._check_shape_(m2)
        batch_axes = self.kwargs['batch_axes']
        keep_axes = self.kwargs['keep_axes']
        sum_axes = self.kwargs['sum_axes']
        left = self.kwargs['left']
        right = self.kwargs['right']

        if verbose:
            print("- %s, shapes=%r @ %r batch_axes=%r keep_axes=%r "
                  "sum_axes=%r" % (
                      self.name, m1.shape, m2.shape, batch_axes, keep_axes, sum_axes))

        if len(m1.shape) != len(m2.shape):
            raise RuntimeError(  # pragma: no cover
                "batch_dot only work with two tensors with the same number "
                "of dimensions not %r @ %r." % (m1.shape, m2.shape))

        dim0 = int(numpy.prod([m1.shape[i] for i in batch_axes]))
        dim0b = int(numpy.prod([m2.shape[i] for i in batch_axes]))
        dimb = int(-1 if keep_axes is None else numpy.prod(
            [m1.shape[i] for i in keep_axes]))
        dim1 = int(numpy.prod([m1.shape[i] for i in sum_axes]))
        dim2 = int(numpy.prod([m2.shape[i] for i in sum_axes]))

        if verbose:
            print(f"- {self.name}, reshape={m1.shape!r} into {dim0, dimb, dim1!r}")
            print(f"- {self.name}, reshape={m2.shape!r} into {dim0b, dimb, dim2!r}")
        m1sh = m1.reshape((dim0, dimb, dim1))
        m2sh = m2.reshape((dim0b, dimb, dim2))

        batch_kind = self.get_dot_kind()
        if batch_kind in ('11', 'N1', 'N1'):
            m1sh = m1sh.reshape((-1, m1sh.shape[-1]))
            m2sh = m2sh.reshape((-1, m2sh.shape[-1]))
            if verbose:
                print("- %s, use gemm with shape %r, %r" % (
                    self.name, m1sh.shape, m2sh.shape))
            dot = gemm_dot(m1sh, m2sh, False, True)
        else:
            dot = m1sh @ numpy.transpose(m2sh, (0, 2, 1))

        # new shape
        new_shape = ([max(m1.shape[i], m2.shape[i]) for i in batch_axes] +
                     [m1.shape[i] for i in left if i not in batch_axes] +
                     [m2.shape[i] for i in right if i not in batch_axes])
        while len(new_shape) < len(m1.shape):
            new_shape.append(1)

        if verbose:
            taken = set(batch_axes) | set(sum_axes)
            ax = [i for i in range(len(m1.shape)) if i not in taken]
            print("- %s, shapes=%r @ %r -> %r" % (
                self.name, m1sh.shape, m2sh.shape, dot.shape))
            print("- %s, batch_axes=%r ax=%r new_shape=%r left=%r right=%r" % (
                self.name, batch_axes, ax, new_shape, left, right))

        output = dot.reshape(tuple(new_shape))
        self._check_shape_(output)
        return output

    def _apply_reduce_sum(self, data, verbose=False, **kwargs):
        self._check_inputs_(1)
        inp = self.inputs[0]
        m = self._get_data(data, inp)
        self._check_shape_(m)
        axes = self.kwargs['axes']
        if verbose:
            print(
                f"- {self.name}, shape={m.shape!r} axes={self.kwargs['axes']!r}")
        output = numpy.sum(m, axis=axes, keepdims=True)
        self._check_shape_(output)
        return output

    def _apply_reduce_sum_mm(self, data, verbose=False, **kwargs):
        self._check_inputs_(2, True)
        inp = self.inputs[0]
        m = self._get_data(data, inp)
        self._check_shape_(m)
        if verbose:
            print(
                f"- {self.name}, shape={m.shape!r} axes={self.kwargs['axes']!r}")
        output = numpy.sum(m, self.kwargs['axes'])
        self._check_shape_(output)
        return output

    def _apply_squeeze(self, data, verbose=False, **kwargs):
        self._check_inputs_(1)
        inp = self.inputs[0]
        m = self._get_data(data, inp)
        axes = self.kwargs['axes']
        if verbose:
            print(
                f"- {self.name}, shape={m.shape!r} axes={self.kwargs['axes']!r}")
        output = m
        for a in axes[::-1]:
            output = numpy.squeeze(output, axis=a)
        return output

    def apply(self, data, verbose=False, **kwargs):
        """
        Applies one operator on the data.

        :param data: dictionary storing the results
        :param verbose: prints out intermediate results
        :param kwargs: additional parameters, see
            methods `_apply*`
        :return: output

        Known additional paramaters:

        * 'matmul_impl': if None calls :epkg:`numpy:einsum` through
          @see fn numpy_extended_dot (default) or 'py' to call
          @see fn numpy_extended_dot_python instead.
        """
        if verbose:
            print()
            print("apply %r  (%s)." % (
                self.name, ", ".join(map(lambda s: str(id(s)), self.inputs))))

        method_name = f"_apply_{self.name}"
        meth = getattr(self, method_name, None)
        if meth is None:
            raise NotImplementedError(  # pragma: no cover
                f"apply not implemented for {self.name!r}.")
        output = meth(data, verbose, **kwargs)

        data[id(self)] = output
        if verbose:
            print("+ %s, shape=%r -- %d" % (self.name, output.shape, id(self)))
        return output

    def _onnx_name(self):
        return 'einsum%d_%s' % (id(self), self.name[:2])

    def _check_onnx_opset_(self, opset, limit):
        if opset is not None and opset < limit:
            raise RuntimeError(  # pragma: no cover
                f"Opset ({opset!r}) must be >= {limit!r} for operator {self.name!r}.")

    def _to_onnx_id(self, names, opset, verbose=False, **kwargs):
        self._check_inputs_(1)
        inp = self.inputs[0]
        name = self._get_data(names, inp)
        yield helper.make_node('Identity', [name], [self._onnx_name()])

    def _to_onnx_expand_dims(self, names, opset, verbose=False, **kwargs):
        self._check_inputs_(1)
        self._check_onnx_opset_(opset, 11)
        inp = self.inputs[0]
        name = self._get_data(names, inp)
        axes = self.kwargs['axes']
        name_axes = name + '_axes'
        yield numpy_helper.from_array(
            numpy.array([a[1] for a in axes], dtype=numpy.int64), name=name_axes)
        s_axes = "".join(map(str, [a[1] for a in axes]))
        yield helper.make_node(
            'Unsqueeze', [name, name_axes], [self._onnx_name()],
            name='Unsqueeze%s_%d' % (s_axes, id(self)))

    def _to_onnx_squeeze(self, names, opset, verbose=False, **kwargs):
        self._check_inputs_(1)
        self._check_onnx_opset_(opset, 11)
        inp = self.inputs[0]
        name = self._get_data(names, inp)
        axes = self.kwargs['axes']
        name_axes = name + '_axes'
        yield numpy_helper.from_array(
            numpy.array(axes, dtype=numpy.int64), name=name_axes)
        s_axes = "".join(map(str, axes))
        yield helper.make_node(
            'Squeeze', [name, name_axes], [self._onnx_name()],
            name='Squeeze%s_%d' % (s_axes, id(self)))

    def _to_onnx_transpose(self, names, opset, verbose=False, **kwargs):
        self._check_inputs_(1)
        inp = self.inputs[0]
        name = self._get_data(names, inp)
        perm = self.kwargs['perm']
        s_perm = "".join(map(str, perm))
        yield helper.make_node(
            'Transpose', [name], [self._onnx_name()], perm=perm,
            name='Transpose%s_%d' % (s_perm, id(self)))

    def _to_onnx_reduce_sum(self, names, opset, verbose=False, **kwargs):
        self._check_inputs_(1)
        self._check_onnx_opset_(opset, 11)
        inp = self.inputs[0]
        name = self._get_data(names, inp)
        axes = self.kwargs['axes']
        name_axes = self._onnx_name() + '_axes'
        yield numpy_helper.from_array(
            numpy.array(axes, dtype=numpy.int64), name=name_axes)
        s_axes = "".join(map(str, axes))
        yield helper.make_node(
            'ReduceSum', [name, name_axes], [self._onnx_name()], keepdims=1,
            name='ReduceSum%s_%d' % (s_axes, id(self)))

    def _to_onnx_mul(self, data, verbose=False, **kwargs):
        self._check_inputs_(2)
        inp1 = self.inputs[0]
        inp2 = self.inputs[1]
        m1 = self._get_data(data, inp1)
        m2 = self._get_data(data, inp2)
        yield helper.make_node('Mul', [m1, m2], [self._onnx_name()])

    def _to_onnx_batch_dot(self, names, opset, verbose=False, **kwargs):  # pylint: disable=R0914
        self._check_inputs_(2)
        self._check_onnx_opset_(opset, 13)
        inp1, inp2 = self.inputs[:2]  # pylint: disable=W0632
        name1 = self._get_data(names, inp1)
        name2 = self._get_data(names, inp2)

        batch_axes = self.kwargs['batch_axes']
        keep_axes = self.kwargs['keep_axes']
        sum_axes = self.kwargs['sum_axes']
        left = self.kwargs['left']
        right = self.kwargs['right']
        root = self._onnx_name()

        def return_name_one():
            name_one = root + "_1"
            return name_one, numpy_helper.from_array(
                numpy.array([1], dtype=numpy.int64), name=name_one)

        name_one = None
        name_shape1 = root + "_shape1"
        name_shape2 = root + "_shape2"
        concat_left = []
        concat_right = []
        yield helper.make_node('Shape', [name1], [name_shape1])
        yield helper.make_node('Shape', [name2], [name_shape2])

        if len(batch_axes) > 0:
            name_batch_axes = root + "_batch_axes"
            yield numpy_helper.from_array(
                numpy.array(batch_axes, dtype=numpy.int64), name=name_batch_axes)

        if len(sum_axes) > 0:
            name_sum_axes = root + "_sum_axes"
            yield numpy_helper.from_array(
                numpy.array(sum_axes, dtype=numpy.int64), name=name_sum_axes)

        # dim0 = int(numpy.prod([m1.shape[i] for i in batch_axes]))
        # dim0b = int(numpy.prod([m2.shape[i] for i in batch_axes]))
        if len(batch_axes) > 1:
            name_dim0 = root + "_dim0"
            name_dim0b = root + "_dim0b"
            name_dim0g = name_dim0 + 'g'
            name_dim0bg = name_dim0b + 'g'
            concat_left.append(name_dim0)
            concat_right.append(name_dim0b)
            yield helper.make_node(
                'Gather', [name_shape1, name_batch_axes], [name_dim0g])
            yield helper.make_node(
                'Gather', [name_shape2, name_batch_axes], [name_dim0bg])
            yield helper.make_node(
                'ReduceProd', [name_dim0g], [name_dim0], keepdims=1)
            yield helper.make_node(
                'ReduceProd', [name_dim0bg], [name_dim0b], keepdims=1)
        elif len(batch_axes) == 1:
            name_dim0g = root + "_dim0g"
            name_dim0bg = root + "_dim0bg"
            name_dim0 = name_dim0g
            name_dim0b = name_dim0bg
            concat_left.append(name_dim0)
            concat_right.append(name_dim0b)
            yield helper.make_node(
                'Gather', [name_shape1, name_batch_axes], [name_dim0g])
            yield helper.make_node(
                'Gather', [name_shape2, name_batch_axes], [name_dim0bg])
        else:
            if name_one is None:
                name_one, cst_init = return_name_one()
                yield cst_init
            name_dim0 = name_one
            name_dim0b = name_one
            concat_left.append(name_dim0)
            concat_right.append(name_dim0b)

        # dimb = int(-1 if keep_axes is None else numpy.prod(
        #     [m1.shape[i] for i in keep_axes]))
        if keep_axes in (-1, None) or len(keep_axes) == 0:
            name_dimb = root + "__1"
            concat_left.append(name_dimb)
            concat_right.append(name_dimb)
            yield numpy_helper.from_array(
                numpy.array([-1], dtype=numpy.int64), name=name_dimb)
        elif len(keep_axes) == 1:
            name_keep_axes = root + "_keep_axes"
            name_dimb = root + "_dimb"
            name_dimbg = name_dimb
            concat_left.append(name_dimb)
            concat_right.append(name_dimb)
            yield numpy_helper.from_array(
                numpy.array(keep_axes, dtype=numpy.int64), name=name_keep_axes)
            yield helper.make_node(
                'Gather', [name_shape1, name_keep_axes], [name_dimbg])
        else:
            name_keep_axes = root + "_keep_axes"
            name_dimb = root + "_dimb"
            name_dimbg = name_dimb + 'g'
            concat_left.append(name_dimb)
            concat_right.append(name_dimb)
            yield numpy_helper.from_array(
                numpy.array(keep_axes, dtype=numpy.int64), name=name_keep_axes)
            yield helper.make_node(
                'Gather', [name_shape1, name_keep_axes], [name_dimbg])
            yield helper.make_node(
                'ReduceProd', [name_dimbg], [name_dimb], keepdims=1)

        # dim1 = int(numpy.prod([m1.shape[i] for i in sum_axes]))
        # dim2 = int(numpy.prod([m2.shape[i] for i in sum_axes]))

        if len(sum_axes) == 0:
            if name_one is None:
                name_one, cst_init = return_name_one()
                yield cst_init
            name_dim1 = name_one
            name_dim2 = name_one
            concat_left.append(name_dim1)
            concat_right.append(name_dim2)
        elif len(sum_axes) == 1:
            name_dim1 = root + "_dim1"
            name_dim2 = root + "_dim2"
            name_dim1g = name_dim1
            name_dim2g = name_dim2
            concat_left.append(name_dim1)
            concat_right.append(name_dim2)
            yield helper.make_node(
                'Gather', [name_shape1, name_sum_axes], [name_dim1g])
            yield helper.make_node(
                'Gather', [name_shape2, name_sum_axes], [name_dim2g])
        else:
            name_dim1 = root + "_dim1"
            name_dim2 = root + "_dim2"
            name_dim1g = name_dim1 + 'g'
            name_dim2g = name_dim2 + 'g'
            concat_left.append(name_dim1)
            concat_right.append(name_dim2)
            yield helper.make_node(
                'Gather', [name_shape1, name_sum_axes], [name_dim1g])
            yield helper.make_node(
                'Gather', [name_shape2, name_sum_axes], [name_dim2g])
            yield helper.make_node(
                'ReduceProd', [name_dim1g], [name_dim1], keepdims=1)
            yield helper.make_node(
                'ReduceProd', [name_dim2g], [name_dim2], keepdims=1)

        batch_kind = self.get_dot_kind()
        if batch_kind in ('11', 'N1', 'N1'):
            # *shape1, *shape2
            name_minus_one = root + "__01"
            yield numpy_helper.from_array(
                numpy.array([-1], dtype=numpy.int64), name=name_minus_one)
            name_agg_shape1_2 = root + f"_resh1_{batch_kind}"
            name_agg_shape2_2 = root + f"_resh2_{batch_kind}"
            yield helper.make_node(
                'Concat', [name_minus_one, name_dim1], [name_agg_shape1_2], axis=0)
            yield helper.make_node(
                'Concat', [name_minus_one, name_dim2], [name_agg_shape2_2], axis=0)

            # m1sh = m1.reshape((-1, dim1))
            # m2sh = m2.reshape((-1, dim2))
            name_agg1_2 = root + "_aresh1"
            name_agg2_2 = root + "_aresh2"
            yield helper.make_node('Reshape', [name1, name_agg_shape1_2], [name_agg1_2])
            yield helper.make_node('Reshape', [name2, name_agg_shape2_2], [name_agg2_2])

            # dot = gemm(m1sh, m2sh, False, True)
            name_dot = root + "_gemm"
            yield helper.make_node(
                'Gemm', [name_agg1_2, name_agg2_2], [name_dot],
                alpha=1., beta=0., transA=0, transB=1)
        else:
            # *shape1, *shape2
            name_agg_shape1 = root + "_resh1"
            name_agg_shape2 = root + "_resh2"
            yield helper.make_node(
                'Concat', concat_left, [name_agg_shape1], axis=0)
            yield helper.make_node(
                'Concat', concat_right, [name_agg_shape2], axis=0)

            # m1sh = m1.reshape((dim0, dimb, dim1))
            # m2sh = m2.reshape((dim0b, dimb, dim2))
            name_agg1 = root + "_aresh1"
            name_agg2 = root + "_aresh2"
            yield helper.make_node('Reshape', [name1, name_agg_shape1], [name_agg1])
            yield helper.make_node('Reshape', [name2, name_agg_shape2], [name_agg2])

            # dot = m1sh @ numpy.transpose(m2sh, (0, 2, 1))
            name_agg2_tr = root + "_aresh2_tr"
            yield helper.make_node(
                'Transpose', [name_agg2], [name_agg2_tr], perm=[0, 2, 1],
                name=f"Transpose021_{id(self)}")

            name_dot = root + "_dot"
            yield helper.make_node(
                'MatMul', [name_agg1, name_agg2_tr], [name_dot])

        # new_shape = ([max(m1.shape[i], m2.shape[i]) for i in batch_axes] +
        #      [m1.shape[i] for i in left if i not in batch_axes] +
        #      [m2.shape[i] for i in right if i not in batch_axes])
        concat_final = []
        if len(batch_axes) > 0:
            name_max_dim = root + "_max_dim"
            concat_final.append(name_max_dim)
            yield helper.make_node(
                'Max', [name_dim0g, name_dim0bg], [name_max_dim])

        left_set = list(sorted(set(left) - (set(batch_axes) & set(left))))
        if len(left_set) > 0:
            name_left_dim = root + "_left_dim"
            name_left_set = root + "_left_set"
            yield numpy_helper.from_array(
                numpy.array(left_set, dtype=numpy.int64), name=name_left_set)
            yield helper.make_node(
                'Gather', [name_shape1, name_left_set], [name_left_dim])
            concat_final.append(name_left_dim)

        right_set = list(sorted(set(right) - (set(batch_axes) & set(right))))
        if len(right_set) > 0:
            name_right_dim = root + "_right_dim"
            name_right_set = root + "_right_set"
            yield numpy_helper.from_array(
                numpy.array(right_set, dtype=numpy.int64), name=name_right_set)
            yield helper.make_node(
                'Gather', [name_shape2, name_right_set], [name_right_dim])
            concat_final.append(name_right_dim)

        name_new_shape = root + '_new_shape'
        diff = (
            self.full_dim -
            (len(batch_axes) + len(left_set) + len(right_set)))
        if diff > 0:
            names_ones = root + "_ones"
            yield numpy_helper.from_array(
                numpy.array([1 for i in range(diff)], dtype=numpy.int64),
                name=names_ones)
            concat_final.append(names_ones)

        yield helper.make_node(
            'Concat', concat_final, [name_new_shape], axis=0)

        name_final = root + '_final'
        yield helper.make_node(
            'Reshape', [name_dot, name_new_shape], [name_final])

    def to_onnx(self, names, opset=None, verbose=False, **kwargs):
        """
        Converts this node into ONNX. Enumerates all ONNX node
        which participate to the conversion. The last one
        is the final output.

        :param names: dictionary where to find already converted name
        :param opset: opset
        :param verbose: prints out intermediate results
        :param kwargs: additional parameter for the conversion
        :return: output
        """
        if opset is None:
            opset = __max_supported_opset__  # pragma: no cover
        if verbose:
            print()
            print("to_onnx %r  (%s) opset=%r." % (
                self.name,
                ", ".join(map(lambda s: str(id(s)), self.inputs)),
                opset))

        method_name = f"_to_onnx_{self.name}"
        meth = getattr(self, method_name, None)
        if meth is None:
            if self.name.endswith("_mm"):
                raise NotImplementedError(
                    "to_onnx not implemented for %r."
                    "You should call method simplify_mm_nodes "
                    "to remove it." % self.name)
            raise NotImplementedError(
                f"to_onnx not implemented for {self.name!r}.")
        for node in meth(names, verbose=verbose, opset=opset, **kwargs):
            if hasattr(node, 'output'):
                names[id(self)] = node.output[0]
                if verbose:
                    print("+ OP %r -- (%s - %d)" %
                          (node.output[0], self.name, id(self)))
            elif verbose:
                # Initializer
                print("+ CT %r -- (%s - %d)" %
                      (node.name, self.name, id(self)))
            yield node

    def get_dot_kind(self):
        """
        Every matrix multiplication can be either:

        * a simple multiplication (`M`) (undetected)
        * a 2D matrix multiplication (`11`)
        * a broadcasted matrix multiplication (`N1` or `1N`)
        * a batch matrix multiplication (`NN`)

        This method returns which kind it is.
        """
        batch_axes = self.kwargs['batch_axes']
        # keep_axes = self.kwargs['keep_axes']
        # sum_axes = self.kwargs['sum_axes']
        # left = self.kwargs['left']
        # right = self.kwargs['right']
        info = self._info
        row_left = info['i_row']
        row_right = info['i_row2']

        batch_left = [row_left[k] for k in batch_axes]
        batch_right = [row_right[k] for k in batch_axes]
        n_left = len(batch_left) > 0 and max(batch_left) == 2
        n_right = len(batch_right) > 0 and max(batch_right) == 2
        return f"{'N' if n_left else '1'}{'N' if n_right else '1'}"


class GraphEinsumSubOp:
    """
    Class gathering all nodes produced to explicit einsum
    operators.

    :param letters: list of distinct letters
    :param mat: matrix, see @see fn analyse_einsum_equation
    :param lengths: lengths of every input
    :param duplicates: see @see fn analyse_einsum_equation
    """

    def __init__(self, letters, mat, lengths, duplicates):
        self._nodes = {}
        self._mark = {}
        self._ops = []
        self._inputs = {}
        self.last_op = None
        self.last_added_op = None
        self.metadata = dict(
            letters=letters, mat=mat, lengths=lengths,
            mat0=mat.copy(), duplicates=duplicates)

    def append(self, op):
        """
        Adds one input or result.

        :param op: integer (an input) or an instance of @see cl EinsumSubOp.
        :return: op or None if op is an integer
        """
        if isinstance(op, int):
            if op in self._nodes:
                raise RuntimeError(  # pragma: no cover
                    "Key %d already added." % op)
            self._nodes[op] = op
            self.last_added_op = op
            self._inputs[op] = op
            return None
        if isinstance(op, EinsumSubOp):
            if op in self._nodes:
                raise RuntimeError(  # pragma: no cover
                    "Key %d already added, op=%r." % (id(op), op))
            self._nodes[id(op)] = op
            self._ops.append(op)
            self.last_added_op = op
            return op
        raise TypeError(  # pragma: no cover
            f"Unexpected type {type(op)!r}.")

    def mark_last_node(self):
        """
        Marks the last node as the final output.
        """
        if self.last_added_op is None:
            raise RuntimeError("last_added_op is None.")  # pragma: no cover
        self.mark(-1, self.last_added_op)

    def mark(self, i, op):
        """
        Marks one input or result as an intermediate result
        after a full einsum step.

        :param op: integer (an input) or an instance of @see cl EinsumSubOp.
        """
        if not isinstance(i, int):
            raise TypeError(  # pragma: no cover
                f"i must an integer not {type(i)!r}.")
        if i != -1 and i not in self._inputs:
            raise RuntimeError(  # pragma: no cover
                "Input %d was not registered in %r." % (i, self._inputs))
        if isinstance(op, EinsumSubOp):
            if id(op) not in self._nodes:
                raise RuntimeError(  # pragma: no cover
                    "Key %d not found, op=%r." % (id(op), op))
            self._mark[i] = op
            self._mark[id(op)] = i
            self.last_op = op
        else:
            raise TypeError(  # pragma: no cover
                f"Unexpected type {type(i)!r}.")

    def __iter__(self):
        "Iterates on nodes."
        for op in self._ops:
            yield op

    def to_dot(self, **kwargs):
        """
        Produces a graph in :epkg:`dot`.

        :param kwargs: additional graph option
        :return: string
        """
        options = {
            'orientation': 'portrait',
            'ranksep': '0.25',
            'nodesep': '0.05',
            'width': '0.5',
            'height': '0.1',
            'size': '5',
            'node': '[shape=record]',
        }
        options.update(kwargs)

        def d2s(d):
            it = []
            for k, v in sorted(d.items()):
                it.append(f"{k}={v}")
            return " ".join(it)

        def d2sd(d):
            it = []
            for k, v in sorted(d.items()):
                if len(v) > 1:
                    it.append(f"{k}={','.join(map(str, v))}")
            return " ".join(it)

        rows = ["digraph{"]
        for k, v in options.items():
            if isinstance(v, str) and "[" in v:
                rows.append(f"{k} {v};")
            else:
                rows.append(f"{k}={v};")
        for k, v in self._nodes.items():
            if isinstance(v, int):
                let = [(r, self.metadata['letters'][i])
                       for i, r in enumerate(self.metadata['mat0'][v])
                       if r != -1]
                dup = self.metadata['duplicates'][v]
                if dup is None:
                    dup = ""
                else:
                    dup = f" - {d2sd(dup)}"
                let.sort()
                letters = "".join(_[1] for _ in let)
                lab = "input %d\\\\n%s\\\\n%s%s" % (
                    v, letters, str(self.metadata['mat0'][v]), dup)
                sk = v
                extended_lab = ""
            else:
                lab = f"{v.name}\\\\n{d2s(v.kwargs)}"
                sk = id(v)
                extended_lab = v.dot_label()
                if extended_lab:
                    extended_lab = "\\\\n" + extended_lab

            if sk in self._mark and isinstance(self._mark[sk], int):
                la = self._mark[sk]
                lab = lab.replace("\\\\n", " - I%d\\\\n" % la)
                s = ('%d [label="%s%s" style=filled '
                     'fillcolor=red];' % (k, lab, extended_lab))
            else:
                s = '%d [label="%s%s"];' % (k, lab, extended_lab)
            rows.append(s)
            if not hasattr(v, 'inputs'):
                continue
            for i in v.inputs:
                vid = i if isinstance(i, int) else id(i)
                s = "%d -> %d;" % (vid, k)
                rows.append(s)
        rows.append("}")
        return "\n".join(rows)

    def apply_sequence(self, *inputs, verbose=False, **kwargs):
        """
        Applies a sequence of operations on a list of inputs.

        :param inputs: inputs:
        :param verbose: prints out intermediate results
        :param kwargs: additional parameters,
            see :meth:`apply
            <mlprodict.testing.einsum.einsum_impl_classes.EinsumSubOp.apply>`.
        :return: output
        """
        if verbose:
            print('######### apply_sequence')
        data = {i: inp for i, inp in enumerate(inputs)}
        last = None
        for op in self:
            last = op.apply(data, verbose=verbose, **kwargs)
        if last is None:
            raise RuntimeError(  # pragma: no cover
                "Sequence of operations is empty.")
        return last

    def clean_unused_nodes(self, verbose=False):
        """
        Cleans nodes with unused outputs.

        :param verbose: display intermediate information
        """

        def iteration(it):
            # Walks through all nodes.
            is_used = {}
            for node in self._ops:
                if not isinstance(node, EinsumSubOp):
                    continue
                if id(node) not in is_used:
                    is_used[id(node)] = []
                for inp in node.inputs:
                    if not isinstance(inp, EinsumSubOp):
                        continue
                    idn = id(inp)
                    if idn not in is_used:
                        is_used[idn] = []
                    is_used[idn].append(id(node))

            # Remove unused nodes.
            removed = []
            for k, v in is_used.items():
                if len(v) == 0:
                    removed.append(k)
            removed = set(removed)
            i_rem = []
            for i, op in enumerate(self._ops):
                if not isinstance(op, EinsumSubOp):
                    continue
                if id(op) in removed and id(op) not in self._mark:
                    i_rem.append((i, id(op)))
            for i, idn in reversed(i_rem):
                if verbose:
                    print("[GraphEinsumSubOp.clean_nodes] remove node "
                          "i=%d: %d - id=%d" % (it, i, idn))
                del self._ops[i]
                del self._nodes[idn]
            return len(i_rem) > 0

        it = 1
        while iteration(it):
            it += 1

        self.last_op = None
        self.last_added_op = None

    def simplify_mm_nodes(self, verbose=False):
        """
        Node name suffixed by `mm` are an artifact to keep
        the graph consistent while building it. They can
        now be replaced by the equivalent node without suffix `mm`.

        :param verbose: display intermediate information
        """
        for op in self:
            if not isinstance(op, EinsumSubOp):
                continue
            if op.name.endswith('_mm'):
                if verbose:
                    print("[GraphEinsumSubOp.simplify_mm_nodes] node %r"
                          " - id=%d" % (op.name, id(op)))
                if len(op.inputs) != 2:
                    raise RuntimeError(  # pragma: no cover
                        "Expecting 2 inputs for node %r not %r id=%r." % (
                            op.name, len(op.inputs), id(op)))
                op.name = op.name[:-3]
                op.inputs = op.inputs[:1]

    def _get_forward_nodes(self):
        """
        Returns the forward nodes.
        """
        forward = {}
        for op in self:
            if isinstance(op, int):
                continue
            for inp in op.inputs:
                key = inp if isinstance(inp, int) else id(inp)
                if key in forward:
                    forward[key].append(op)
                else:
                    forward[key] = [op]
        return forward

    def _pprint_forward(self):
        rows = []
        for op in self:
            line = "%r <- %s(%s)" % (
                id(op), op.name,
                ", ".join(map(str, [id(_) for _ in op.inputs])))
            rows.append(line)
        return "\n".join(rows)

    def _replace_node_sequence(self, added, deleted):
        """
        Removes a sequence of nodes. The method does not check
        that the graph remains consistent.
        """
        forward = self._get_forward_nodes()
        key = id(deleted[-1])
        if key not in forward:
            raise RuntimeError(  # pragma: no cover
                "Key {} missing in all forward nodes (other keys {}), "
                "all keys:\n{}".format(
                    key, [id(_) for _ in deleted],
                    self._pprint_forward()))

        # deletion
        mark_input = None
        for d in deleted:
            del self._nodes[id(d)]
            if id(d) in self._mark:
                del self._mark[id(d)]
                dels = []
                for k, v in self._mark.items():
                    if id(v) == id(d):
                        mark_input = k
                        dels.append(k)
                if len(dels) != 1:
                    raise RuntimeError(  # pragma: no cover
                        "Input %d has more than one marked operator "
                        "(%r)." % (id(d), dels))
                del self._mark[dels[0]]

        dels = set(id(o) for o in deleted)
        rem = []
        for i, op in enumerate(self._ops):
            if id(op) in dels:
                rem.append(i)
        if len(rem) != len(deleted):
            raise RuntimeError(  # pragma: no cover
                f"Mismatched length {rem!r}, {dels!r}, len={len(deleted)!r}.")
        for i in reversed(rem):
            del self._ops[i]
        self.last_add_op = None

        # insertion
        if added is not None:
            self._ops.insert(rem[0], added)
            self._nodes[id(added)] = added
            for op in forward[key]:
                new_inputs = list(op.inputs)
                for i in range(len(op.inputs)):  # pylint: disable=C0200
                    if id(op.inputs[i]) == key:
                        new_inputs[i] = added
                op.inputs = tuple(new_inputs)
            if mark_input is not None:
                self.mark(mark_input, added)
        else:
            inps = deleted[0].inputs
            if len(inps) != 1:
                raise RuntimeError(  # pragma: no cover
                    "More than one input. Call another method.")
            inp = inps[0]
            for op in forward[key]:
                new_inputs = list(op.inputs)
                for i in range(len(op.inputs)):  # pylint: disable=C0200
                    if id(op.inputs[i]) == key:
                        new_inputs[i] = inp
                op.inputs = tuple(new_inputs)
            if mark_input is not None:
                self.mark(mark_input, inp)

    def remove_duplicate_transpose(self, verbose=False):
        """
        Removes consecutive transpose by merging them.

        :param verbose: display intermediate information
        """
        modif = 1
        while modif > 0:
            modif = 0
            candidates = []
            forward = self._get_forward_nodes()
            for op in self:
                if op.name == "transpose":
                    inp = op.inputs[0]
                    if (isinstance(inp, EinsumSubOp) and
                            inp.name == 'transpose' and
                            len(forward[id(inp)]) == 1):
                        candidates.append(op)

            if len(candidates) > 0:
                modif = 1
                # Not efficient to take the first one and to
                # start again but the graph should not be too big.
                cand = candidates[0]
                op2 = cand
                op1 = cand.inputs[0]
                perm1 = op1.kwargs['perm']
                perm2 = op2.kwargs['perm']
                if len(perm1) != len(perm2):
                    raise RuntimeError(  # pragma: no cover
                        "Transposition should have the same length "
                        "%r, %r." % (perm1, perm2))
                perm = list(perm1)
                for i in range(len(perm)):  # pylint: disable=C0200
                    perm[i] = perm1[perm2[i]]
                if list(range(len(perm))) == perm:
                    # identity, everything needs to be removed
                    new_op = None
                else:
                    new_op = op2.__class__(
                        op2.full_dim, op2.name, op1.inputs[0],
                        perm=tuple(perm))
                self._replace_node_sequence(new_op, [op1, op2])
                if verbose:
                    print(  # pragma: no cover
                        "[GraphEinsumSubOp.remove_duplicate_transpose] remove nodes %r"
                        " - id=%d,%d + %d perm1=%r perm2=%r -> perm=%r" % (
                            op2.name, id(op1), id(op2),
                            id(new_op) if new_op is not None else -1,
                            perm1, perm2, perm))

    def to_onnx(self, output, *inputs, dtype=None, verbose=False,
                opset=None, **kwargs):
        """
        Converts the graph into ONNX.

        :param output: output name
        :param inputs: input names
        :param dtype: type used for all operators
        :param opset: desired opset, None for the last one
        :param verbose: display intermediate operators
        :param kwargs: additional parameter to use when building
            the ONNX graph, list of supported parameters:
            *name*, *ir_version*, *producer_name*,
            *producer_version*, *initializer*
        :return: ONNX graph

        Not all graphs can be converted into ONNX. Only graphs produced
        with `strategy='numpy'` can be converted otherwise the following
        error shows up:

        ::

            NotImplementedError: to_onnx not implemented for 'matmul'.
        """
        from ...onnx_tools.optim import onnx_remove_node_unused

        # inputs
        if opset is None:
            opset = __max_supported_opset__
        if verbose:
            print("[GraphEinsumSubOp.to_onnx] %r -> %s opset=%r "
                  "dtype=%r" % (inputs, output, opset, dtype))
        onx_inputs = []
        proto = guess_proto_dtype(
            numpy.float32 if dtype is None else dtype)
        lengths = self.metadata['lengths']
        names = {}
        for inp, le in zip(inputs, lengths):
            if isinstance(inp, tuple):
                name, typ = inp
                if le != len(typ.shape):
                    raise ValueError(  # pragma: no cover
                        "Irreconcialable shapes for input %r: "
                        "%r != len(%r)." % (name, le, typ.shape))
                proto = guess_proto_dtype(guess_numpy_type(typ))
                onx_inputs.append(
                    helper.make_tensor_value_info(name, proto, typ.shape))
                names[len(names)] = name
            else:
                onx_inputs.append(
                    helper.make_tensor_value_info(
                        inp, proto, [None for i in range(le)]))
                names[len(names)] = inp

        # output
        onx_output = helper.make_tensor_value_info(
            output, proto, [None for i in range(lengths[-1])])

        # nodes
        nodes = []
        inits = []
        if "initializer" in kwargs:
            inits.extend(kwargs['initializer'])
        for op in self:
            for onx_node in op.to_onnx(names, verbose=verbose, opset=opset):
                if hasattr(onx_node, 'output'):
                    nodes.append(onx_node)
                else:
                    inits.append(onx_node)

        # last node
        last_node = nodes[-1]
        nodes.append(helper.make_node(
            'Identity', [last_node.output[0]], [output]))

        # Builds the graph
        model = helper.make_model(
            opset_imports=[helper.make_operatorsetid('', opset)],
            ir_version=kwargs.get('ir_version', get_ir_version(opset)),
            producer_name=kwargs.get('producer_name', 'mlprodict'),
            producer_version=kwargs.get('producer_version', "0.0.dev"),
            graph=helper.make_graph(
                name=kwargs.get('name', 'einsum'),
                inputs=onx_inputs, outputs=[onx_output],
                initializer=inits, nodes=nodes))

        return onnx_remove_node_unused(model)
